LANGUAGE_HELP = ''
REPOSITORY_HELP = (
    'This element is used to give the managing custodian of the material '
    'being described.')
SCRIPT_HELP = ''

# XML-related constants.
EAD_NAMESPACE = 'http://ead3.archivists.org/schema/'
EAD_NS = '{%s}' % EAD_NAMESPACE
NS_MAP = {'e': EAD_NAMESPACE}

# Attribute groups, being a restatement of the attribute mixins and
# attributes from models.py.
ACCESS_ATTRS = ['identifier', 'rules', 'source']
BASE_ATTRS = ['base']
COMMON_ATTRS = ['altrender', 'audience', 'id']
COVERAGE_ATTRS = ['coverage']
DATE_ATTRS = ['calendar', 'certainty', 'era']
ENCODING_ANALOG_ATTRS = ['encodinganalog']
LABEL_ATTRS = ['label']
LANG_SCRIPT_ATTRS = ['lang', 'script']
LINK_ATTRS = ['actuate', 'arcrole', 'href', 'linkrole', 'linktitle', 'show']
LOCAL_TYPE_ATTRS = ['localtype']
NORMAL_ATTRS = ['normal']
RELATED_ENCODING_ATTRS = ['relatedencoding']
RELATOR_ATTRS = ['relator']
TRANSLITERATION_ATTRS = ['transliteration']
UNIT_DATE_ATTRS = DATE_ATTRS + ['datechar', 'unitdatetype']
XPOINTER_ATTRS = ['xpointer']

# Attributes per element. Note that some attributes need to be handled
# specially (eg, parent) and so are not listed here.
ABBR_ATTRS = COMMON_ATTRS + LANG_SCRIPT_ATTRS + ['expan']
ABSTRACT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
ACCESSRESTRICT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
ACCRUALS_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
ACQINFO_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
ADDRESS_ATTRS = COMMON_ATTRS + LANG_SCRIPT_ATTRS
ADDRESSLINE_ATTRS = COMMON_ATTRS + LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
AGENCYCODE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
AGENCYNAME_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
AGENT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS
AGENTTYPE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    ['value']
ALTFORMAVAIL_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
APPRAISAL_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
ARRANGEMENT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
ARCHDESC_ATTRS = BASE_ATTRS + COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS + RELATED_ENCODING_ATTRS + \
    ['level', 'otherlevel']
AUTHOR_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
BIBLIOGRAPHY_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
BIOGHIST_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
CITATION_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LINK_ATTRS + ['lastdatetimeverified']
CONTAINER_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS + ['containerid']
CONTROL_ATTRS = BASE_ATTRS + COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + RELATED_ENCODING_ATTRS + [
        'countryencoding', 'dateencoding', 'langencoding',
        'repositoryencoding', 'scriptencoding']
CONTROLACCESS_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
CONTROLNOTE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
CONVENTIONDECLARATION_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
CUSTODHIST_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
DAO_ATTRS = COMMON_ATTRS + COVERAGE_ATTRS + ENCODING_ANALOG_ATTRS + \
    LABEL_ATTRS + LANG_SCRIPT_ATTRS + LINK_ATTRS + LOCAL_TYPE_ATTRS + \
    XPOINTER_ATTRS + ['daotype', 'identifier', 'otherdaotype']
DAOSET_ATTRS = BASE_ATTRS + COMMON_ATTRS + COVERAGE_ATTRS + \
    ENCODING_ANALOG_ATTRS + LABEL_ATTRS + LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
DATERANGE_ATTRS = COMMON_ATTRS + LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
DATESET_ATTRS = COMMON_ATTRS + LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
DATESINGLE_ATTRS = COMMON_ATTRS + LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS + \
    ['notafter', 'notbefore', 'standarddate']
DESCRIPTIVENOTE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS
DID_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS
DIDNOTE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
DIMENSIONS_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS + ['unit']
EAD_ATTRS = BASE_ATTRS + COMMON_ATTRS + LANG_SCRIPT_ATTRS + \
    RELATED_ENCODING_ATTRS
EDITIONSTMT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS
EVENTDATETIME_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + ['standarddatetime']
EVENTDESCRIPTION_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
EVENTTYPE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    ['value']
FILEDESC_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS
FILEPLAN_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
FROMTODATE_ATTRS = DATESINGLE_ATTRS
GEOGNAME_ATTRS = ACCESS_ATTRS + COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS + NORMAL_ATTRS + RELATOR_ATTRS
HEAD_ATTRS = COMMON_ATTRS + LANG_SCRIPT_ATTRS + ['althead']
INDEX_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
LANGMATERIAL_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS
LANGUAGE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS
LANGUAGEDECLARATION_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS
LANGUAGESET_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS
LEGALSTATUS_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
LOCALCONTROL_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
LOCALTYPEDECLARATION_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS
MAINTENANCEAGENCY_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + ['countrycode']
MAINTENANCEEVENT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS
MAINTENANCEHISTORY_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS
MAINTENANCESTATUS_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + ['value']
MATERIALSPEC_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
NAME_ATTRS = ACCESS_ATTRS + COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS + NORMAL_ATTRS + RELATOR_ATTRS
NOTESTMT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS
OBJECTXMLWRAP_ATTRS = COMMON_ATTRS + LANG_SCRIPT_ATTRS
ODD_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
ORIGINALSLOC_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
ORIGINATION_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
OTHERAGENCYCODE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
OTHERFINDAID_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
OTHERRECORDID_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
PART_ATTRS = ACCESS_ATTRS + COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
PHYSDESC_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
PHYSDESCSET_ATTRS = COMMON_ATTRS + COVERAGE_ATTRS + ENCODING_ANALOG_ATTRS + \
    LABEL_ATTRS + LANG_SCRIPT_ATTRS + ['parallel']
PHYSDESCSTRUCTURED_ATTRS = COMMON_ATTRS + COVERAGE_ATTRS + \
    ENCODING_ANALOG_ATTRS + LABEL_ATTRS + LANG_SCRIPT_ATTRS + \
    ['otherphysdescstructuredtype', 'physdescstructuredtype']
PHYSFACET_ATTRS = ACCESS_ATTRS + COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
PHYSLOC_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
PHYSTECH_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
PREFERCITE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
PROCESSINFO_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
PUBLICATIONSTATUS_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + ['value']
PUBLICATIONSTMT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS
QUANTITY_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    ['approximate']
RECORDID_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    ['instanceurl']
RELATEDMATERIAL_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
RELATION_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LINK_ATTRS + ['lastdatetimeverified', 'otherrelationtype', 'relationtype']
RELATIONENTRY_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS + TRANSLITERATION_ATTRS
RELATIONS_ATTRS = BASE_ATTRS + COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
REPOSITORY_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
REPRESENTATION_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LINK_ATTRS + LOCAL_TYPE_ATTRS
RIGHTSDECLARATION_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
SCOPECONTENT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
SCRIPT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS
SEPARATEDMATERIAL_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
SERIESSTMT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS
SOURCE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LINK_ATTRS + ['lastdatetimeverified']
SOURCEENTRY_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + TRANSLITERATION_ATTRS
SOURCES_ATTRS = BASE_ATTRS + COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS
SPONSOR_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
SUBTITLE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LOCAL_TYPE_ATTRS
TERM_ATTRS = ACCESS_ATTRS + COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + TRANSLITERATION_ATTRS + ['lastdatetimeverified']
TITLEPROPER_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS + ['render']
TITLESTMT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS
UNITDATE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LANG_SCRIPT_ATTRS + \
    LABEL_ATTRS + NORMAL_ATTRS + UNIT_DATE_ATTRS
UNITDATESTRUCTURED_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LABEL_ATTRS + UNIT_DATE_ATTRS
UNITID_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS + \
    ['countrycode', 'identifier', 'repositorycode']
UNITTITLE_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + LABEL_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS + NORMAL_ATTRS
UNITTYPE_ATTRS = ACCESS_ATTRS + COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS
USERESTRICT_ATTRS = COMMON_ATTRS + ENCODING_ANALOG_ATTRS + \
    LANG_SCRIPT_ATTRS + LOCAL_TYPE_ATTRS



import re

from django.core.exceptions import ValidationError


NAME_START_CHAR = ''.join(r'''
  :
  A-Z
  a-z
  _
  \u00C0-\u00D6
  \u00D8-\u00F6
  \u00F8-\u02FF
  \u0370-\u037D
  \u037F-\u1FFF
  \u200C-\u200D
  \u2070-\u218F
  \u2C00-\u2FEF
  \u3001-\uD7FF
  \uF900-\uFDCF
  \uFDF0-\uFFFD
  \u10000-\uEFFFF
'''.split())
NAME_CHAR = ''.join(r'''
  {}
  \-
  \.
  0-9
  \u00B7
  \u0300-\u036F
  \u203F-\u2040
'''.format(NAME_START_CHAR).split())

DATE_TIME_PATTERN = re.compile(
    r'\d{4}(-\d{2}(-\d{2}(T\d{2}:\d{2}:\d{2}(Z|([+-]\d{2}:\d{2}))?)?)?)?')
ID_PATTERN = re.compile(
    r'[{}]([{}])*'.format(NAME_START_CHAR, NAME_CHAR))
NM_TOKEN_PATTERN = re.compile(r'[{}]+'.format(NAME_CHAR))


def validate_date_time(value):
    """Raise ValidationError if `value` does not constitute an ISO 8601
    pattern.

    The valid patterns are:

    * YYYY-MM-DD
    * YYYY-MM
    * YYYY
    * YYYY-MM-DDThh:mm:ss
    * YYYY-MM-DDThh:mm:ss[+|-]hh:mm
    * YYYY-MM-DDThh:mm:ssZ

    """
    if DATE_TIME_PATTERN.fullmatch(value) is None:
        raise ValidationError('"{}" is not a valid date-time'.format(value))


def validate_id(value):
    """Raise ValidationError if `value` does not consitute an ID.

    An ID must begin with an alpha, not numeric, character, either
    upper or lowercase, and may contain a . (period), - (hyphen), or _
    (underscore), but not a blank space.

    """
    if ID_PATTERN.fullmatch(value) is None:
        raise ValidationError('"{}" is not a valid ID'.format(value))


def validate_nmtoken(value):
    """Raise ValidationError if `value` does not constitute a name token.

    A name token can consists of any alpha or numeric character, as
    well as a . (period), : (colon), - (hyphen), or _ (underscore),
    but not a blank space.

    """
    if NM_TOKEN_PATTERN.fullmatch(value) is None:
        raise ValidationError('"{}" is not a valid name'.format(value))


def validate_token(value):
    """Raise ValidationError if `value` does not constitute a token.

    A token is a type of string that may not contain carriage return,
    line feed or tab characters, leading or trailing spaces, nor any
    internal sequence of two or more spaces.

    """
    if value.strip() != value or '\r' in value or '\n' in value or \
       '\t' in value or '  ' in value:
        raise ValidationError('"{}" is not a valid token'.format(value))



"""Import records from supplied EAD XML files.

Validates imported files against the EAD3 XSD, and also against the
Django models' validators, partly due to small differences between the
two (eg, an empty physfacet is valid in the XML, not in the Django
model) and partly as a sanity check.

"""

import logging
import os.path

from django.core.management.base import BaseCommand, CommandError
from django.db import transaction

from lxml import etree

from controlled_vocabulary.utils import search_term

import reversion

from ...constants import (
    ABBR_ATTRS, ABSTRACT_ATTRS, ACCESSRESTRICT_ATTRS, ACCRUALS_ATTRS,
    ACQINFO_ATTRS, ADDRESS_ATTRS, ADDRESSLINE_ATTRS, AGENCYCODE_ATTRS,
    AGENCYNAME_ATTRS, AGENT_ATTRS, AGENTTYPE_ATTRS, ALTFORMAVAIL_ATTRS,
    APPRAISAL_ATTRS, ARCHDESC_ATTRS, ARRANGEMENT_ATTRS, AUTHOR_ATTRS,
    BIBLIOGRAPHY_ATTRS, BIOGHIST_ATTRS, CITATION_ATTRS, CONTAINER_ATTRS,
    CONTROL_ATTRS, CONTROLACCESS_ATTRS, CONTROLNOTE_ATTRS,
    CONVENTIONDECLARATION_ATTRS, CUSTODHIST_ATTRS, DAO_ATTRS, DAOSET_ATTRS,
    DATERANGE_ATTRS, DATESET_ATTRS, DATESINGLE_ATTRS, DESCRIPTIVENOTE_ATTRS,
    DID_ATTRS, DIDNOTE_ATTRS, DIMENSIONS_ATTRS, EAD_ATTRS, EAD_NS,
    EDITIONSTMT_ATTRS, EVENTDATETIME_ATTRS, EVENTDESCRIPTION_ATTRS,
    EVENTTYPE_ATTRS, FILEDESC_ATTRS, FILEPLAN_ATTRS, FROMTODATE_ATTRS,
    GEOGNAME_ATTRS, HEAD_ATTRS, INDEX_ATTRS, LANGMATERIAL_ATTRS,
    LANGUAGE_ATTRS, LANGUAGEDECLARATION_ATTRS, LANGUAGESET_ATTRS,
    LEGALSTATUS_ATTRS, LOCALCONTROL_ATTRS, LOCALTYPEDECLARATION_ATTRS,
    MAINTENANCEAGENCY_ATTRS, MAINTENANCEEVENT_ATTRS, MAINTENANCEHISTORY_ATTRS,
    MAINTENANCESTATUS_ATTRS, MATERIALSPEC_ATTRS, NAME_ATTRS, NOTESTMT_ATTRS,
    NS_MAP, OBJECTXMLWRAP_ATTRS, ODD_ATTRS, ORIGINALSLOC_ATTRS,
    ORIGINATION_ATTRS, OTHERAGENCYCODE_ATTRS, OTHERFINDAID_ATTRS,
    OTHERRECORDID_ATTRS, PART_ATTRS, PHYSDESC_ATTRS, PHYSDESCSET_ATTRS,
    PHYSDESCSTRUCTURED_ATTRS, PHYSFACET_ATTRS, PHYSLOC_ATTRS, PHYSTECH_ATTRS,
    PREFERCITE_ATTRS, PROCESSINFO_ATTRS, PUBLICATIONSTATUS_ATTRS,
    PUBLICATIONSTMT_ATTRS, QUANTITY_ATTRS, RECORDID_ATTRS,
    RELATEDMATERIAL_ATTRS, RELATION_ATTRS, RELATIONENTRY_ATTRS,
    RELATIONS_ATTRS, REPOSITORY_ATTRS, REPRESENTATION_ATTRS,
    RIGHTSDECLARATION_ATTRS, SCOPECONTENT_ATTRS, SCRIPT_ATTRS,
    SEPARATEDMATERIAL_ATTRS, SERIESSTMT_ATTRS, SOURCE_ATTRS, SOURCEENTRY_ATTRS,
    SOURCES_ATTRS, SPONSOR_ATTRS, SUBTITLE_ATTRS, TERM_ATTRS,
    TITLEPROPER_ATTRS, TITLESTMT_ATTRS, UNITDATE_ATTRS,
    UNITDATESTRUCTURED_ATTRS, UNITID_ATTRS, UNITTITLE_ATTRS, UNITTYPE_ATTRS,
    USERESTRICT_ATTRS)
from ...models import (
    Abstract, AccessRestrict, Accruals, AcqInfo, AgencyName, AltFormAvail,
    Appraisal, Arrangement, Author, Bibliography, BiogHist, Container,
    ControlAccess, ControlNote, ConventionDeclaration, CustodHist, DAOSet,
    DAOSetDAO, DIdDAO, DIdNote, DIdPhysDescStructured,
    DIdPhysDescStructuredDimensions, DIdPhysDescStructuredPhysFacet, EAD,
    EventDescription, FilePlan, Index, LangMaterial, LangMaterialLanguage,
    LanguageDeclaration, LanguageSet, LanguageSetLanguage, LanguageSetScript,
    LegalStatus, LocalControl, LocalTypeDeclaration, MaintenanceEvent,
    MaterialSpec, ODD, OriginalsLoc, Origination, OriginationCorpName,
    OriginationCorpNamePart, OriginationFamName, OriginationFamNamePart,
    OriginationName, OriginationNamePart, OriginationPersName,
    OriginationPersNamePart, OtherAgencyCode, OtherFindAid, OtherRecordID,
    PhysDesc, PhysDescSet, PhysDescSetPhysDescStructured,
    PhysDescSetPhyDescStructuredDimensions,
    PhysDescSetPhysDescStructuredPhysFacet, PhysLoc, PhysTech, PreferCite,
    ProcessInfo, RelatedMaterial, Relation, RelationDateRange,
    RelationDateSingle, RelationEntry, Repository, RepositoryAddressLine,
    RepositoryCorpName, RepositoryCorpNamePart, RepositoryFamName,
    RepositoryFamNamePart, RepositoryName, RepositoryNamePart,
    RepositoryPersName, RepositoryPersNamePart, Representation,
    RightsDeclaration, ScopeContent, SeparatedMaterial, Source, SourceEntry,
    Sponsor, Subtitle, TitleProper, UnitDate, UnitDateStructured,
    UnitDateStructuredDateRange, UnitDateStructuredDateSingle, UnitId,
    UnitTitle, UseRestrict)


# Parser options.
HELP = ('Import archival data from EAC XML files. Creates initial revisions '
        'on successful import.')
PATHS_HELP = 'Path to EAD XML file to import.'

EAD_XSD_FILENAME = 'ead3.xsd'

# Logging and error messages.
EMPTY_NAME_MSG = 'Name to be imported within file at {} is empty.'
NO_MATCHING_LANGUAGE_MSG = 'Could not find language matching code "{}".'
NO_MATCHING_SCRIPT_MSG = 'Could not find script matching code "{}".'
NOT_WELL_FORMED_XML_MSG = (
    'Document "{}" is either not well-formed XML or is invalid: {}')
RECORD_EXISTS_MSG = (
    'Document "{}" has recordid "{}" that already exists, sourced from {}')
UNSUPPORTED_RECORD_LEVEL_MSG = (
    'Document "{}" specifies unsupported archdesc/@level "{}"')


def import_attrs(source, attrs, dest, prefix=''):
    """Add any of `attrs` on `source` to `dest`, appending `prefix` to the
    fieldname."""
    for attr in attrs:
        attr_value = source.get(attr)
        if attr == 'id':
            attr = 'ead_id'
        if attr_value is not None:
            if attr == 'lang':
                lang_code = attr_value
                attr_value = search_term('iso639-2', lang_code, exact=True)
                if attr_value is None:
                    raise CommandError(NO_MATCHING_LANGUAGE_MSG.format(
                        lang_code))
            elif attr == 'script':
                script_code = attr_value
                attr_value = search_term('iso15924', script_code, exact=True)
                if attr_value is None:
                    raise CommandError(NO_MATCHING_SCRIPT_MSG.format(
                        script_code))
            elif attr in ('approximate', 'parallel'):
                # BooleanFields.
                if attr_value == 'true':
                    attr_value = True
                else:
                    attr_value = False
            setattr(dest, prefix + attr, attr_value)


def serialise(element):
    """Return `element` serialised, minus the tags for `element`."""
    parts = []
    parts.append(element.text or '')
    for child in element:
        parts.append(etree.tostring(
            child, encoding='unicode', xml_declaration=False))
    return ''.join(parts)


def xpath(source, expression):
    return source.xpath(expression, namespaces=NS_MAP)


class Command(BaseCommand):

    help = HELP
    logger = logging.getLogger(__name__)

    def add_arguments(self, parser):
        parser.add_argument('paths', help=PATHS_HELP, metavar='FILE',
                            nargs='+', type=str)

    @transaction.atomic
    def handle(self, *args, **options):
        self._record_ids = {}
        xsd = self._get_xsd()
        parser = etree.XMLParser(ns_clean=True, remove_blank_text=True,
                                 schema=xsd)
        for path in options['paths']:
            self._path = os.path.abspath(path)
            self._handle_path(self._path, parser)

    def _get_xsd(self):
        xsd_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                EAD_XSD_FILENAME)
        xsd_doc = etree.parse(xsd_path)
        return etree.XMLSchema(xsd_doc)

    def _handle_path(self, path, parser):
        try:
            tree = etree.parse(path, parser=parser)
        except etree.XMLSyntaxError as e:
            raise CommandError(NOT_WELL_FORMED_XML_MSG.format(path, e))
        ead = tree.getroot()
        try:
            self.stdout.write('Importing file at {}.'.format(path))
            self._import_ead(ead)
        except Exception as e:
            self.stdout.write('Problem occurred importing file at {}.'.format(
                path))
            raise e

    def _import_abbr(self, abbr, parent_obj):
        """abbr may occur as the child of many elements; this method is only
        for abbr as a child of elements where it may occur only once."""
        parent_obj.abbr = abbr.text
        import_attrs(abbr, ABBR_ATTRS, parent_obj, 'abbr_')
        parent_obj.save()

    def _import_abstract(self, abstract, ead_obj):
        abstract_obj = Abstract(did=ead_obj, abstract=serialise(abstract))
        import_attrs(abstract, ABSTRACT_ATTRS, abstract_obj)
        abstract_obj.full_clean()
        abstract_obj.save()

    def _import_accessrestrict(self, accessrestrict, ead_obj):
        accessrestrict_obj = AccessRestrict(
            archdesc=ead_obj, accessrestrict=serialise(accessrestrict))
        import_attrs(accessrestrict, ACCESSRESTRICT_ATTRS, accessrestrict_obj)
        accessrestrict_obj.full_clean()
        accessrestrict_obj.save()

    def _import_accruals(self, accruals, ead_obj):
        accruals_obj = Accruals(archdesc=ead_obj, accruals=serialise(accruals))
        import_attrs(accruals, ACCRUALS_ATTRS, accruals_obj)
        accruals_obj.full_clean()
        accruals_obj.save()

    def _import_acqinfo(self, acqinfo, ead_obj):
        acqinfo_obj = AcqInfo(archdesc=ead_obj, acqinfo=serialise(acqinfo))
        import_attrs(acqinfo, ACQINFO_ATTRS, acqinfo_obj)
        acqinfo_obj.full_clean()
        acqinfo_obj.save()

    def _import_address(self, address, repository_obj):
        import_attrs(address, ADDRESS_ATTRS, repository_obj, 'address_')
        repository_obj.full_clean()
        repository_obj.save()
        for addressline in xpath(address, 'e:addressline'):
            self._import_addressline(addressline, repository_obj)

    def _import_addressline(self, addressline, repository_obj):
        addressline_obj = RepositoryAddressLine(
            address=repository_obj, addressline=serialise(addressline))
        import_attrs(addressline, ADDRESSLINE_ATTRS, addressline_obj)
        addressline_obj.full_clean()
        addressline_obj.save()

    def _import_agencycode(self, agencycode, ead_obj):
        import_attrs(agencycode, AGENCYCODE_ATTRS, ead_obj, 'agencycode_')
        ead_obj.agencycode = agencycode.text

    def _import_agencyname(self, agencyname, ead_obj):
        agencyname_obj = AgencyName(
            maintenanceagency=ead_obj, agencyname=agencyname.text)
        import_attrs(agencyname, AGENCYNAME_ATTRS, agencyname_obj)
        agencyname_obj.full_clean()
        agencyname_obj.save()

    def _import_agent(self, agent, maintenanceevent_obj):
        import_attrs(agent, AGENT_ATTRS, maintenanceevent_obj, 'agent_')
        if agent.text:
            maintenanceevent_obj.agent = agent.text

    def _import_agenttype(self, agenttype, maintenanceevent_obj):
        import_attrs(agenttype, AGENTTYPE_ATTRS, maintenanceevent_obj,
                     'agenttype_')
        if agenttype.text:
            maintenanceevent_obj.agenttype = agenttype.text

    def _import_altformavail(self, altformavail, ead_obj):
        altformavail_obj = AltFormAvail(
            archdesc=ead_obj, altformavail=serialise(altformavail))
        import_attrs(altformavail, ALTFORMAVAIL_ATTRS, altformavail_obj)
        altformavail_obj.full_clean()
        altformavail_obj.save()

    def _import_appraisal(self, appraisal, ead_obj):
        appraisal_obj = Appraisal(
            archdesc=ead_obj, appraisal=serialise(appraisal))
        import_attrs(appraisal, APPRAISAL_ATTRS, appraisal_obj)
        appraisal_obj.full_clean()
        appraisal_obj.save()

    def _import_archdesc(self, archdesc, ead_obj):
        import_attrs(archdesc, ARCHDESC_ATTRS, ead_obj, 'archdesc_')
        did = archdesc[0]
        self._import_did(did, ead_obj)
        for accessrestrict in xpath(archdesc, 'e:accessrestrict'):
            self._import_accessrestrict(accessrestrict, ead_obj)
        for accruals in xpath(archdesc, 'e:accruals'):
            self._import_accruals(accruals, ead_obj)
        for acqinfo in xpath(archdesc, 'e:acqinfo'):
            self._import_acqinfo(acqinfo, ead_obj)
        for altformavail in xpath(archdesc, 'e:altformavail'):
            self._import_altformavail(altformavail, ead_obj)
        for appraisal in xpath(archdesc, 'e:appraisal'):
            self._import_appraisal(appraisal, ead_obj)
        for arrangement in xpath(archdesc, 'e:arrangement'):
            self._import_arrangement(arrangement, ead_obj)
        for bibliography in xpath(archdesc, 'e:bibliography'):
            self._import_bibliography(bibliography, ead_obj)
        for bioghist in xpath(archdesc, 'e:bioghist'):
            self._import_bioghist(bioghist, ead_obj)
        for controlaccess in xpath(archdesc, 'e:controlaccess'):
            self._import_controlaccess(controlaccess, ead_obj)
        for custodhist in xpath(archdesc, 'e:custodhist'):
            self._import_custodhist(custodhist, ead_obj)
        for fileplan in xpath(archdesc, 'e:fileplan'):
            self._import_fileplan(fileplan, ead_obj)
        for index in xpath(archdesc, 'e:index'):
            self._import_index(index, ead_obj)
        for legalstatus in xpath(archdesc, 'e:legalstatus'):
            self._import_legalstatus(legalstatus, ead_obj)
        for odd in xpath(archdesc, 'e:odd'):
            self._import_odd(odd, ead_obj)
        for originalsloc in xpath(archdesc, 'e:originalsloc'):
            self._import_originalsloc(originalsloc, ead_obj)
        for otherfindaid in xpath(archdesc, 'e:otherfindaid'):
            self._import_otherfindaid(otherfindaid, ead_obj)
        for phystech in xpath(archdesc, 'e:phystech'):
            self._import_phystech(phystech, ead_obj)
        for prefercite in xpath(archdesc, 'e:prefercite'):
            self._import_prefercite(prefercite, ead_obj)
        for processinfo in xpath(archdesc, 'e:processinfo'):
            self._import_processinfo(processinfo, ead_obj)
        for relatedmaterial in xpath(archdesc, 'e:relatedmaterial'):
            self._import_relatedmaterial(relatedmaterial, ead_obj)
        for relations in xpath(archdesc, 'e:relations'):
            self._import_relations(relations, ead_obj)
        for scopecontent in xpath(archdesc, 'e:scopecontent'):
            self._import_scopecontent(scopecontent, ead_obj)
        for separatedmaterial in xpath(archdesc, 'e:separatedmaterial'):
            self._import_separatedmaterial(separatedmaterial, ead_obj)
        for userestrict in xpath(archdesc, 'e:userestrict'):
            self._import_userestrict(userestrict, ead_obj)

    def _import_arrangement(self, arrangement, ead_obj):
        arrangement_obj = Arrangement(
            archdesc=ead_obj, arrangement=serialise(arrangement))
        import_attrs(arrangement, ARRANGEMENT_ATTRS, arrangement_obj)
        arrangement_obj.full_clean()
        arrangement_obj.save()

    def _import_author(self, author, ead_obj):
        author_obj = Author(titlestmt=ead_obj, author=serialise(author))
        import_attrs(author, AUTHOR_ATTRS, author_obj)
        author_obj.full_clean()
        author_obj.save()

    def _import_bibliography(self, bibliography, ead_obj):
        bibliography_obj = Bibliography(
            archdesc=ead_obj, bibliography=serialise(bibliography))
        import_attrs(bibliography, BIBLIOGRAPHY_ATTRS, bibliography_obj)
        bibliography_obj.full_clean()
        bibliography_obj.save()

    def _import_bioghist(self, bioghist, ead_obj):
        bioghist_obj = BiogHist(archdesc=ead_obj, bioghist=serialise(bioghist))
        import_attrs(bioghist, BIOGHIST_ATTRS, bioghist_obj)
        bioghist_obj.full_clean()
        bioghist_obj.save()

    def _import_citation(self, citation, parent_obj):
        parent_obj.citation = serialise(citation)
        import_attrs(citation, CITATION_ATTRS, parent_obj, 'citation_')

    def _import_container(self, container, ead_obj):
        container_obj = Container(did=ead_obj, container=serialise(container))
        import_attrs(container, CONTAINER_ATTRS, container_obj)
        container_obj.full_clean()
        container_obj.save()
        return container_obj

    def _import_container_parents(self, container, container_obj, ead_obj):
        for idref in container.get('parent', '').split():
            parent_obj = Container.objects.get(did=ead_obj, ead_id=idref)
            container_obj.parents.add(parent_obj)

    def _import_control(self, control, ead_obj):
        import_attrs(control, CONTROL_ATTRS, ead_obj, 'control_')
        self._import_recordid(control[0], ead_obj)
        ead_obj.save()
        for otherrecordid in xpath(control, 'e:otherrecordid'):
            self._import_otherrecordid(otherrecordid, ead_obj)
        for representation in xpath(control, 'e:representation'):
            self._import_representation(representation, ead_obj)
        filedesc = xpath(control, 'e:filedesc')[0]
        self._import_filedesc(filedesc, ead_obj)
        maintenancestatus = xpath(control, 'e:maintenancestatus')[0]
        self._import_maintenancestatus(maintenancestatus, ead_obj)
        for publicationstatus in xpath(control, 'e:publicationstatus'):
            self._import_publicationstatus(publicationstatus, ead_obj)
        maintenanceagency = xpath(control, 'e:maintenanceagency')[0]
        self._import_maintenanceagency(maintenanceagency, ead_obj)
        for languagedeclaration in xpath(control, 'e:languagedeclaration'):
            self._import_languagedeclaration(languagedeclaration, ead_obj)
        for conventiondeclaration in xpath(control, 'e:conventiondeclaration'):
            self._import_conventiondeclaration(conventiondeclaration, ead_obj)
        for rightsdeclaration in xpath(control, 'e:rightsdeclaration'):
            self._import_rightsdeclaration(rightsdeclaration, ead_obj)
        for localtypedeclaration in xpath(control, 'e:localtypedeclaration'):
            self._import_localtypedeclaration(localtypedeclaration, ead_obj)
        for localcontrol in xpath(control, 'e:localcontrol'):
            self._import_localcontrol(localcontrol, ead_obj)
        maintenancehistory = xpath(control, 'e:maintenancehistory')[0]
        self._import_maintenancehistory(maintenancehistory, ead_obj)
        for sources in xpath(control, 'e:sources'):
            self._import_sources(sources, ead_obj)

    def _import_controlaccess(self, controlaccess, ead_obj):
        controlaccess_obj = ControlAccess(
            archdesc=ead_obj, controlaccess=serialise(controlaccess))
        import_attrs(controlaccess, CONTROLACCESS_ATTRS, controlaccess_obj)
        controlaccess_obj.full_clean()
        controlaccess_obj.save()

    def _import_controlnote(self, controlnote, ead_obj):
        controlnote_obj = ControlNote(
            notestmt=ead_obj, controlnote=serialise(controlnote))
        import_attrs(controlnote, CONTROLNOTE_ATTRS, controlnote_obj)
        controlnote_obj.full_clean()
        controlnote_obj.save()

    def _import_conventiondeclaration(self, conventiondeclaration, ead_obj):
        conventiondeclaration_obj = ConventionDeclaration(
            control=ead_obj)
        import_attrs(conventiondeclaration, CONVENTIONDECLARATION_ATTRS,
                     conventiondeclaration_obj)
        for abbr in xpath(conventiondeclaration, 'e:abbr'):
            self._import_abbr(abbr, conventiondeclaration_obj)
        for citation in xpath(conventiondeclaration, 'e:citation'):
            self._import_citation(citation, conventiondeclaration_obj)
        for descriptivenote in xpath(
                conventiondeclaration, 'e:descriptivenote'):
            self._import_descriptivenote(
                descriptivenote, conventiondeclaration_obj)
        conventiondeclaration_obj.full_clean()
        conventiondeclaration_obj.save()

    def _import_custodhist(self, custodhist, ead_obj):
        custodhist_obj = CustodHist(
            archdesc=ead_obj, custodhist=serialise(custodhist))
        import_attrs(custodhist, CUSTODHIST_ATTRS, custodhist_obj)
        custodhist_obj.full_clean()
        custodhist_obj.save()

    def _import_dao(self, dao, parent_obj):
        if dao.getparent().tag == EAD_NS + 'did':
            dao_model = DIdDAO
        else:
            dao_model = DAOSetDAO
        dao_obj = dao_model(parent=parent_obj)
        import_attrs(dao, DAO_ATTRS, dao_obj)
        for descriptivenote in xpath(dao, 'e:descriptivenote'):
            self._import_descriptivenote(descriptivenote, dao_obj)
        dao_obj.full_clean()
        dao_obj.save()

    def _import_daoset(self, daoset, ead_obj):
        daoset_obj = DAOSet(did=ead_obj)
        import_attrs(daoset, DAOSET_ATTRS, daoset_obj)
        for descriptivenote in xpath(daoset, 'e:descriptivenote'):
            self._import_descriptivenote(descriptivenote, daoset_obj)
        daoset_obj.full_clean()
        daoset_obj.save()
        for dao in xpath(daoset, 'e:dao'):
            self._import_dao(dao, daoset_obj)

    def _import_daterange(self, daterange, parent_obj, daterange_model):
        daterange_obj = daterange_model(parent=parent_obj)
        import_attrs(daterange, DATERANGE_ATTRS, daterange_obj)
        for fromdate in xpath(daterange, 'e:fromdate'):
            self._import_fromtodate(fromdate, daterange_obj, 'fromdate')
        for todate in xpath(daterange, 'e:todate'):
            self._import_fromtodate(todate, daterange_obj, 'todate')
        daterange_obj.full_clean()
        daterange_obj.save()

    def _import_datesingle(self, datesingle, parent_obj, datesingle_model):
        datesingle_obj = datesingle_model(parent=parent_obj)
        import_attrs(datesingle, DATESINGLE_ATTRS, datesingle_obj)
        if datesingle.text or len(datesingle):
            datesingle_obj.datesingle = serialise(datesingle)
        datesingle_obj.full_clean()
        datesingle_obj.save()

    def _import_descriptivenote(self, descriptivenote, parent_obj, prefix=''):
        setattr(parent_obj, prefix + 'descriptivenote', serialise(
            descriptivenote))
        import_attrs(descriptivenote, DESCRIPTIVENOTE_ATTRS, parent_obj,
                     prefix + 'descriptivenote_')

    def _import_did(self, did, ead_obj):
        import_attrs(did, DID_ATTRS, ead_obj, 'did_')
        for head in xpath(did, 'e:head'):
            self._import_head(head, ead_obj, 'did_')
        for abstract in xpath(did, 'e:abstract'):
            self._import_abstract(abstract, ead_obj)
        containers = xpath(did, 'e:container')
        container_objs = []
        for container in containers:
            container_objs.append(self._import_container(container, ead_obj))
        # container @parent requires processing after all Containers
        # have been created.
        for container, container_obj in zip(containers, container_objs):
            self._import_container_parents(container, container_obj, ead_obj)
        for dao in xpath(did, 'e:dao'):
            self._import_dao(dao, ead_obj)
        for daoset in xpath(did, 'e:daoset'):
            self._import_daoset(daoset, ead_obj)
        for didnote in xpath(did, 'e:didnote'):
            self._import_didnote(didnote, ead_obj)
        for langmaterial in xpath(did, 'e:langmaterial'):
            self._import_langmaterial(langmaterial, ead_obj)
        for materialspec in xpath(did, 'e:materialspec'):
            self._import_materialspec(materialspec, ead_obj)
        for origination in xpath(did, 'e:origination'):
            self._import_origination(origination, ead_obj)
        for physdesc in xpath(did, 'e:physdesc'):
            self._import_physdesc(physdesc, ead_obj)
        for physdescset in xpath(did, 'e:physdescset'):
            self._import_physdescset(physdescset, ead_obj)
        for physdescstructured in xpath(did, 'e:physdescstructured'):
            self._import_physdescstructured(physdescstructured, ead_obj)
        physlocs = xpath(did, 'e:physloc')
        physloc_objs = []
        for physloc in physlocs:
            physloc_objs.append(self._import_physloc(physloc, ead_obj))
        # physloc @parent requires processing after all
        # PhysLocs have been created.
        for physloc, physloc_obj in zip(physlocs, physloc_objs):
            self._import_physloc_parents(physloc, physloc_obj, ead_obj)
        for repository in xpath(did, 'e:repository'):
            self._import_repository(repository, ead_obj)
        for unitdate in xpath(did, 'e:unitdate'):
            self._import_unitdate(unitdate, ead_obj)
        for unitdatestructured in xpath(did, 'e:unitdatestructured'):
            self._import_unitdatestructured(unitdatestructured, ead_obj)
        for unitid in xpath(did, 'e:unitid'):
            self._import_unitid(unitid, ead_obj)
        for unittitle in xpath(did, 'e:unittitle'):
            self._import_unittitle(unittitle, ead_obj)

    def _import_didnote(self, didnote, ead_obj):
        didnote_obj = DIdNote(did=ead_obj, didnote=serialise(didnote))
        import_attrs(didnote, DIDNOTE_ATTRS, didnote_obj)
        didnote_obj.full_clean()
        didnote_obj.save()

    def _import_dimensions(self, dimensions, pds_obj, dimensions_model):
        dimensions_obj = dimensions_model(
            physdescstructured=pds_obj, dimensions=serialise(dimensions))
        import_attrs(dimensions, DIMENSIONS_ATTRS, dimensions_obj)
        dimensions_obj.full_clean()
        dimensions_obj.save()

    def _import_ead(self, ead):
        """Create and return an EAD object from the data in `ead`.

        The record is saved. If a record with the same record_id
        already exists, an exception is raised.

        """
        recordid = xpath(ead, 'e:control/e:recordid')[0].text
        try:
            ead_obj = EAD.objects.get(recordid=recordid)
            source = self._record_ids.get(recordid, 'an earlier import')
            raise CommandError(RECORD_EXISTS_MSG.format(
                self._path, recordid, source))
        except EAD.DoesNotExist:
            pass
        self._record_ids[recordid] = self._path
        with reversion.create_revision():
            ead_obj = EAD()
            import_attrs(ead, EAD_ATTRS, ead_obj)
            self._import_control(ead[0], ead_obj)
            self._import_archdesc(ead[1], ead_obj)
            ead_obj.full_clean()
            ead_obj.save()
            reversion.set_comment('Created record from EAD3 XML.')
        return ead_obj

    def _import_editionstmt(self, editionstmt, ead_obj):
        import_attrs(editionstmt, EDITIONSTMT_ATTRS, ead_obj,
                     'editionstmt_')
        ead_obj.editionstmt = serialise(editionstmt)

    def _import_eventdatetime(self, eventdatetime, maintenanceevent_obj):
        import_attrs(eventdatetime, EVENTDATETIME_ATTRS, maintenanceevent_obj,
                     'eventdatetime_')
        if eventdatetime.text:
            maintenanceevent_obj.eventdatetime = eventdatetime.text

    def _import_eventdescription(self, eventdescription, maintenanceevent_obj):
        eventdescription_obj = EventDescription(
            maintenanceevent=maintenanceevent_obj)
        import_attrs(eventdescription, EVENTDESCRIPTION_ATTRS,
                     eventdescription_obj)
        eventdescription_obj.eventdescription = eventdescription.text
        eventdescription_obj.full_clean()
        eventdescription_obj.save()

    def _import_eventtype(self, eventtype, maintenanceevent_obj):
        import_attrs(eventtype, EVENTTYPE_ATTRS, maintenanceevent_obj,
                     'eventtype_')
        if eventtype.text:
            maintenanceevent_obj.eventtype = eventtype.text

    def _import_filedesc(self, filedesc, ead_obj):
        import_attrs(filedesc, FILEDESC_ATTRS, ead_obj, 'filedesc_')
        self._import_titlestmt(xpath(filedesc, 'e:titlestmt')[0], ead_obj)
        for editionstmt in xpath(filedesc, 'e:editionstmt'):
            self._import_editionstmt(editionstmt, ead_obj)
        for publicationstmt in xpath(filedesc, 'e:publicationstmt'):
            self._import_publicationstmt(publicationstmt, ead_obj)
        for seriesstmt in xpath(filedesc, 'e:seriesstmt'):
            self._import_seriesstmt(seriesstmt, ead_obj)
        for notestmt in xpath(filedesc, 'e:notestmt'):
            self._import_notestmt(notestmt, ead_obj)

    def _import_fileplan(self, fileplan, ead_obj):
        fileplan_obj = FilePlan(archdesc=ead_obj, fileplan=serialise(fileplan))
        import_attrs(fileplan, FILEPLAN_ATTRS, fileplan_obj)
        fileplan_obj.full_clean()
        fileplan_obj.save()

    def _import_fromtodate(self, fromtodate, daterange_obj, field_name):
        # Due to the usefulness of some attributes on fromdate/todate
        # for processing, store the attributes in model fields, along
        # with the serialised element.
        if fromtodate.text or len(fromtodate):
            setattr(daterange_obj, field_name, serialise(fromtodate))
        import_attrs(fromtodate, FROMTODATE_ATTRS, daterange_obj,
                     field_name + '_')

    def _import_geogname(self, geogname, relation_obj):
        relation_obj.geogname = serialise(geogname)
        import_attrs(geogname, GEOGNAME_ATTRS, relation_obj, 'geogname_')

    def _import_head(self, head, parent_obj, prefix=''):
        setattr(parent_obj, prefix + 'head', serialise(head))
        import_attrs(head, HEAD_ATTRS, parent_obj, prefix + 'head_')

    def _import_index(self, index, ead_obj):
        index_obj = Index(archdesc=ead_obj, index=serialise(index))
        import_attrs(index, INDEX_ATTRS, index_obj)
        index_obj.full_clean()
        index_obj.save()

    def _import_langmaterial(self, langmaterial, ead_obj):
        langmaterial_obj = LangMaterial(did=ead_obj)
        import_attrs(langmaterial, LANGMATERIAL_ATTRS, langmaterial_obj)
        for descriptivenote in xpath(langmaterial, 'e:descriptivenote'):
            self._import_descriptivenote(descriptivenote, langmaterial_obj)
        langmaterial_obj.full_clean()
        langmaterial_obj.save()
        for language in xpath(langmaterial, 'e:language'):
            self._import_language(
                language, langmaterial_obj, LangMaterialLanguage)
        for languageset in xpath(langmaterial, 'e:languageset'):
            self._import_languageset(languageset, langmaterial_obj)

    def _import_language(self, language, parent_obj, language_model):
        language_obj = language_model(parent=parent_obj)
        import_attrs(language, LANGUAGE_ATTRS, language_obj)
        langcode = language.get('langcode')
        if langcode:
            langcode_obj = search_term('iso639-2', langcode)
            language_obj.langcode = langcode_obj
        language_obj.language = language.text
        language_obj.full_clean()
        language_obj.save()

    def _import_languagedeclaration(self, ld, ead_obj):
        language = ld[0]
        script = ld[1]
        langcode = search_term('iso639-2', language.get('langcode'))
        scriptcode = search_term('iso15924', script.get('scriptcode'))
        ld_obj = LanguageDeclaration(control=ead_obj)
        import_attrs(ld, LANGUAGEDECLARATION_ATTRS, ld_obj)
        import_attrs(language, LANGUAGE_ATTRS, ld_obj, 'language_')
        import_attrs(script, SCRIPT_ATTRS, ld_obj, 'script_el_')
        ld_obj.language = language.text
        ld_obj.language_langcode = langcode
        ld_obj.script_el = script.text
        ld_obj.script_el_scriptcode = scriptcode
        for descriptivenote in xpath(ld, 'e:descriptivenote'):
            self._import_descriptivenote(descriptivenote, ld_obj)
        ld_obj.full_clean()
        ld_obj.save()

    def _import_languageset(self, languageset, langmaterial_obj):
        languageset_obj = LanguageSet(langmaterial=langmaterial_obj)
        import_attrs(languageset, LANGUAGESET_ATTRS, languageset_obj)
        for descriptivenote in xpath(languageset, 'e:descriptivenote'):
            self._import_descriptivenote(descriptivenote, languageset_obj)
        languageset_obj.full_clean()
        languageset_obj.save()
        for language in xpath(languageset, 'e:language'):
            self._import_language(
                language, languageset_obj, LanguageSetLanguage)
        for script in xpath(languageset, 'e:script'):
            self._import_script(script, languageset_obj)

    def _import_legalstatus(self, legalstatus, ead_obj):
        legalstatus_obj = LegalStatus(
            archdesc=ead_obj, legalstatus=serialise(legalstatus))
        import_attrs(legalstatus, LEGALSTATUS_ATTRS, legalstatus_obj)
        legalstatus_obj.full_clean()
        legalstatus_obj.save()

    def _import_localcontrol(self, localcontrol, ead_obj):
        localcontrol_obj = LocalControl(control=ead_obj)
        import_attrs(localcontrol, LOCALCONTROL_ATTRS, localcontrol_obj)
        for term in xpath(localcontrol, 'e:term'):
            self._import_term(term, localcontrol_obj)
        for daterange in xpath(localcontrol, 'e:daterange'):
            import_attrs(daterange, DATERANGE_ATTRS, localcontrol_obj,
                         'daterange_')
            for fromdate in xpath(daterange, 'e:fromdate'):
                self._import_fromtodate(fromdate, localcontrol_obj, 'fromdate')
            for todate in xpath(daterange, 'e:todate'):
                self._import_fromtodate(todate, localcontrol_obj, 'todate')
        for datesingle in xpath(localcontrol, 'e:datesingle'):
            import_attrs(datesingle, DATESINGLE_ATTRS, localcontrol_obj,
                         'datesingle_')
            if datesingle.text or len(datesingle):
                localcontrol_obj.datesingle = serialise(datesingle)
        localcontrol_obj.full_clean()
        localcontrol_obj.save()

    def _import_localtypedeclaration(self, ltd, ead_obj):
        ltd_obj = LocalTypeDeclaration(control=ead_obj)
        import_attrs(ltd, LOCALTYPEDECLARATION_ATTRS, ltd_obj)
        for abbr in xpath(ltd, 'e:abbr'):
            self._import_abbr(abbr, ltd_obj)
        for citation in xpath(ltd, 'e:citation'):
            self._import_citation(citation, ltd_obj)
        for descriptivenote in xpath(ltd, 'e:descriptivenote'):
            self._import_descriptivenote(descriptivenote, ltd_obj)
        ltd_obj.full_clean()
        ltd_obj.save()

    def _import_maintenanceagency(self, maintenanceagency, ead_obj):
        import_attrs(maintenanceagency, MAINTENANCEAGENCY_ATTRS, ead_obj,
                     'maintenanceagency_')
        for agencycode in xpath(maintenanceagency, 'e:agencycode'):
            self._import_agencycode(agencycode, ead_obj)
        for otheragencycode in xpath(maintenanceagency, 'e:otheragencycode'):
            self._import_otheragencycode(otheragencycode, ead_obj)
        for agencyname in xpath(maintenanceagency, 'e:agencyname'):
            self._import_agencyname(agencyname, ead_obj)
        for descriptivenote in xpath(maintenanceagency, 'e:descriptivenote'):
            self._import_descriptivenote(
                descriptivenote, ead_obj, 'maintenanceagency_')

    def _import_maintenanceevent(self, maintenanceevent, ead_obj):
        maintenanceevent_obj = MaintenanceEvent(maintenancehistory=ead_obj)
        import_attrs(maintenanceevent, MAINTENANCEEVENT_ATTRS,
                     maintenanceevent_obj)
        self._import_eventtype(maintenanceevent[0], maintenanceevent_obj)
        eventdatetime = maintenanceevent[1]
        self._import_eventdatetime(eventdatetime, maintenanceevent_obj)
        agenttype = maintenanceevent[2]
        self._import_agenttype(agenttype, maintenanceevent_obj)
        agent = maintenanceevent[3]
        self._import_agent(agent, maintenanceevent_obj)
        maintenanceevent_obj.full_clean()
        maintenanceevent_obj.save()
        for eventdescription in maintenanceevent[4:]:
            self._import_eventdescription(
                eventdescription, maintenanceevent_obj)

    def _import_maintenancehistory(self, maintenancehistory, ead_obj):
        import_attrs(maintenancehistory, MAINTENANCEHISTORY_ATTRS, ead_obj,
                     'maintenancehistory_')
        for maintenanceevent in maintenancehistory:
            self._import_maintenanceevent(maintenanceevent, ead_obj)

    def _import_maintenancestatus(self, maintenancestatus, ead_obj):
        import_attrs(maintenancestatus, MAINTENANCESTATUS_ATTRS, ead_obj,
                     'maintenancestatus_')
        if maintenancestatus.text:
            ead_obj.maintenancestatus = maintenancestatus.text

    def _import_materialspec(self, materialspec, ead_obj):
        materialspec_obj = MaterialSpec(
            did=ead_obj, materialspec=serialise(materialspec))
        import_attrs(materialspec, MATERIALSPEC_ATTRS, materialspec_obj)
        materialspec_obj.full_clean()
        materialspec_obj.save()

    def _import_name(self, name, name_model, part_model, parent_obj, field):
        """Import name from `name` element, using `name_model` and
        `part_model`, and return it."""
        assembled_name = ''.join(xpath(name, 'e:part//text()'))
        if not assembled_name:
            raise CommandError(EMPTY_NAME_MSG.format(self._path))
        name_obj = name_model()
        setattr(name_obj, field, parent_obj)
        import_attrs(name, NAME_ATTRS, name_obj)
        name_obj.full_clean()
        name_obj.save()
        for idx, part in enumerate(name):
            self._import_part(part, name_obj, part_model, idx)
        name_obj.full_clean()
        name_obj.save()  # To set the assembled_name from the parts.
        return name_obj

    def _import_notestmt(self, notestmt, ead_obj):
        import_attrs(notestmt, NOTESTMT_ATTRS, ead_obj, 'notestmt_')
        for controlnote in notestmt:
            self._import_controlnote(controlnote, ead_obj)

    def _import_objectxmlwrap(self, objectxmlwrap, parent_obj):
        import_attrs(objectxmlwrap, OBJECTXMLWRAP_ATTRS, parent_obj,
                     'objectxmlwrap_')
        parent_obj.objectxmlwrap = serialise(objectxmlwrap)

    def _import_odd(self, odd, ead_obj):
        odd_obj = ODD(archdesc=ead_obj, odd=serialise(odd))
        import_attrs(odd, ODD_ATTRS, odd_obj)
        odd_obj.full_clean()
        odd_obj.save()

    def _import_originalsloc(self, originalsloc, ead_obj):
        originalsloc_obj = OriginalsLoc(
            archdesc=ead_obj, originalsloc=serialise(originalsloc))
        import_attrs(originalsloc, ORIGINALSLOC_ATTRS, originalsloc_obj)
        originalsloc_obj.full_clean()
        originalsloc_obj.save()

    def _import_origination(self, origination, ead_obj):
        origination_obj = Origination(did=ead_obj)
        import_attrs(origination, ORIGINATION_ATTRS, origination_obj)
        origination_obj.full_clean()
        origination_obj.save()
        for corpname in xpath(origination, 'e:corpname'):
            self._import_name(
                corpname, OriginationCorpName, OriginationCorpNamePart,
                origination_obj, 'origination')
        for famname in xpath(origination, 'e:famname'):
            self._import_name(
                famname, OriginationFamName, OriginationFamNamePart,
                origination_obj, 'origination')
        for name in xpath(origination, 'e:name'):
            self._import_name(name, OriginationName, OriginationNamePart,
                              origination_obj, 'origination')
        for persname in xpath(origination, 'e:persname'):
            self._import_name(
                persname, OriginationPersName, OriginationPersNamePart,
                origination_obj, 'origination')
        origination_obj.full_clean()

    def _import_otheragencycode(self, otheragencycode, ead_obj):
        otheragencycode_obj = OtherAgencyCode(
            maintenanceagency=ead_obj, otheragencycode=otheragencycode.text)
        import_attrs(otheragencycode, OTHERAGENCYCODE_ATTRS,
                     otheragencycode_obj)
        otheragencycode_obj.full_clean()
        otheragencycode_obj.save()

    def _import_otherfindaid(self, otherfindaid, ead_obj):
        otherfindaid_obj = OtherFindAid(
            archdesc=ead_obj, otherfindaid=serialise(otherfindaid))
        import_attrs(otherfindaid, OTHERFINDAID_ATTRS, otherfindaid_obj)
        otherfindaid_obj.full_clean()
        otherfindaid_obj.save()

    def _import_otherrecordid(self, otherrecordid, ead_obj):
        otherrecordid_obj = OtherRecordID(
            control=ead_obj, otherrecordid=otherrecordid.text)
        import_attrs(otherrecordid, OTHERRECORDID_ATTRS, otherrecordid_obj)
        otherrecordid_obj.full_clean()
        otherrecordid_obj.save()

    def _import_part(self, part, name_obj, part_model, position):
        part_obj = part_model(
            name=name_obj, part=serialise(part), order=position+1)
        import_attrs(part, PART_ATTRS, part_obj)
        part_obj.full_clean()
        part_obj.save()

    def _import_physdesc(self, physdesc, ead_obj):
        physdesc_obj = PhysDesc(did=ead_obj, physdesc=serialise(physdesc))
        import_attrs(physdesc, PHYSDESC_ATTRS, physdesc_obj)
        physdesc_obj.full_clean()
        physdesc_obj.save()

    def _import_physdescset(self, physdescset, ead_obj):
        physdescset_obj = PhysDescSet(did=ead_obj)
        import_attrs(physdescset, PHYSDESCSET_ATTRS, physdescset_obj)
        physdescset_obj.full_clean()
        physdescset_obj.save()
        for pds in xpath(physdescset, 'e:physdescstructured'):
            self._import_physdescstructured(pds, physdescset_obj)

    def _import_physdescstructured(self, pds, parent_obj):
        if pds.getparent().tag == EAD_NS + 'did':
            pds_model = DIdPhysDescStructured
            physfacet_model = DIdPhysDescStructuredPhysFacet
            dimensions_model = DIdPhysDescStructuredDimensions
        else:
            pds_model = PhysDescSetPhysDescStructured
            physfacet_model = PhysDescSetPhysDescStructuredPhysFacet
            dimensions_model = PhysDescSetPhyDescStructuredDimensions
        pds_obj = pds_model(parent=parent_obj)
        import_attrs(pds, PHYSDESCSTRUCTURED_ATTRS, pds_obj)
        for quantity in xpath(pds, 'e:quantity'):
            self._import_quantity(quantity, pds_obj)
        for unittype in xpath(pds, 'e:unittype'):
            self._import_unittype(unittype, pds_obj)
        for descriptivenote in xpath(pds, 'e:descriptivenote'):
            self._import_descriptivenote(descriptivenote, pds_obj)
        pds_obj.full_clean()
        pds_obj.save()
        for dimensions in xpath(pds, 'e:dimensions'):
            self._import_dimensions(dimensions, pds_obj, dimensions_model)
        for physfacet in xpath(pds, 'e:physfacet'):
            self._import_physfacet(physfacet, pds_obj, physfacet_model)

    def _import_physfacet(self, physfacet, pds_obj, physfacet_model):
        physfacet_obj = physfacet_model(
            physdescstructured=pds_obj, physfacet=serialise(physfacet))
        import_attrs(physfacet, PHYSFACET_ATTRS, physfacet_obj)
        physfacet_obj.full_clean()
        physfacet_obj.save()

    def _import_physloc(self, physloc, ead_obj):
        physloc_obj = PhysLoc(did=ead_obj, physloc=serialise(physloc))
        import_attrs(physloc, PHYSLOC_ATTRS, physloc_obj)
        physloc_obj.full_clean()
        physloc_obj.save()
        return physloc_obj

    def _import_physloc_parents(self, physloc, physloc_obj, ead_obj):
        for idref in physloc.get('parent', '').split():
            parent_obj = PhysLoc.objects.get(did=ead_obj, ead_id=idref)
            physloc_obj.parents.add(parent_obj)

    def _import_phystech(self, phystech, ead_obj):
        phystech_obj = PhysTech(archdesc=ead_obj, phystech=serialise(phystech))
        import_attrs(phystech, PHYSTECH_ATTRS, phystech_obj)
        phystech_obj.full_clean()
        phystech_obj.save()

    def _import_prefercite(self, prefercite, ead_obj):
        prefercite_obj = PreferCite(
            archdesc=ead_obj, prefercite=serialise(prefercite))
        import_attrs(prefercite, PREFERCITE_ATTRS, prefercite_obj)
        prefercite_obj.full_clean()
        prefercite_obj.save()

    def _import_processinfo(self, processinfo, ead_obj):
        processinfo_obj = ProcessInfo(
            archdesc=ead_obj, processinfo=serialise(processinfo))
        import_attrs(processinfo, PROCESSINFO_ATTRS, processinfo_obj)
        processinfo_obj.full_clean()
        processinfo_obj.save()

    def _import_publicationstatus(self, publicationstatus, ead_obj):
        import_attrs(publicationstatus, PUBLICATIONSTATUS_ATTRS, ead_obj,
                     'publicationstatus_')
        if publicationstatus.text:
            ead_obj.publicationstatus = publicationstatus.text

    def _import_publicationstmt(self, publicationstmt, ead_obj):
        import_attrs(publicationstmt, PUBLICATIONSTMT_ATTRS, ead_obj,
                     'publicationstmt_')
        ead_obj.publicationstmt = serialise(publicationstmt)

    def _import_quantity(self, quantity, physdescstructured_obj):
        physdescstructured_obj.quantity = quantity.text
        import_attrs(quantity, QUANTITY_ATTRS, physdescstructured_obj,
                     'quantity_')

    def _import_recordid(self, recordid, ead_obj):
        ead_obj.recordid = recordid.text
        import_attrs(recordid, RECORDID_ATTRS, ead_obj, 'recordid_')

    def _import_relatedmaterial(self, rm, ead_obj):
        rm_obj = RelatedMaterial(
            archdesc=ead_obj, relatedmaterial=serialise(rm))
        import_attrs(rm, RELATEDMATERIAL_ATTRS, rm_obj)
        rm_obj.full_clean()
        rm_obj.save()

    def _import_relation(self, relation, ead_obj):
        relation_obj = Relation(relations=ead_obj)
        import_attrs(relation, RELATION_ATTRS, relation_obj)
        relation_obj.save()
        for relationentry in xpath(relation, 'e:relationentry'):
            self._import_relationentry(relationentry, relation_obj)
        for objectxmlwrap in xpath(relation, 'e:objectxmlwrap'):
            self._import_objectxmlwrap(objectxmlwrap, relation_obj)
        for daterange in xpath(relation, './/e:daterange'):
            self._import_daterange(daterange, relation_obj, RelationDateRange)
        for dateset in xpath(relation, 'e:dateset'):
            import_attrs(dateset, DATESET_ATTRS, relation_obj, 'dateset_')
        for datesingle in xpath(relation, './/e:datesingle'):
            self._import_datesingle(
                datesingle, relation_obj, RelationDateSingle)
        for geogname in xpath(relation, 'e:geogname'):
            self._import_geogname(geogname, relation_obj)
        for descriptivenote in xpath(relation, 'e:descriptivenote'):
            self._import_descriptivenote(descriptivenote, relation_obj)
        relation_obj.full_clean()
        relation_obj.save()

    def _import_relations(self, relations, ead_obj):
        import_attrs(relations, RELATIONS_ATTRS, ead_obj, 'relations_')
        for relation in relations:
            self._import_relation(relation, ead_obj)

    def _import_relationentry(self, relationentry, relation_obj):
        relationentry_obj = RelationEntry(relation=relation_obj)
        import_attrs(relationentry, RELATIONENTRY_ATTRS, relationentry_obj)
        relationentry_obj.relationentry = relationentry.text
        relationentry_obj.full_clean()
        relationentry_obj.save()

    def _import_repository(self, repository, ead_obj):
        repository_obj = Repository(did=ead_obj)
        import_attrs(repository, REPOSITORY_ATTRS, repository_obj)
        repository_obj.full_clean()
        repository_obj.save()
        for corpname in xpath(repository, 'e:corpname'):
            self._import_name(
                corpname, RepositoryCorpName, RepositoryCorpNamePart,
                repository_obj, 'repository')
        for famname in xpath(repository, 'e:famname'):
            self._import_name(
                famname, RepositoryFamName, RepositoryFamNamePart,
                repository_obj, 'repository')
        for name in xpath(repository, 'e:name'):
            self._import_name(name, RepositoryName, RepositoryNamePart,
                              repository_obj, 'repository')
        for persname in xpath(repository, 'e:persname'):
            self._import_name(
                persname, RepositoryPersName, RepositoryPersNamePart,
                repository_obj, 'repository')
        for address in xpath(repository, 'e:address'):
            self._import_address(address, repository_obj)
        repository_obj.full_clean()
        repository_obj.save()  # To set the assembled_name from the names.

    def _import_representation(self, representation, ead_obj):
        representation_obj = Representation(control=ead_obj)
        import_attrs(representation, REPRESENTATION_ATTRS, representation_obj)
        if representation.text:
            representation_obj.representation = representation.text
        representation_obj.full_clean()
        representation_obj.save()

    def _import_rightsdeclaration(self, rightsdeclaration, ead_obj):
        rd_obj = RightsDeclaration(control=ead_obj)
        import_attrs(rightsdeclaration, RIGHTSDECLARATION_ATTRS, rd_obj)
        for abbr in xpath(rightsdeclaration, 'e:abbr'):
            self._import_abbr(abbr, rd_obj)
        for citation in xpath(rightsdeclaration, 'e:citation'):
            self._import_citation(citation, rd_obj)
        for descriptivenote in xpath(rightsdeclaration, 'e:descriptivenote'):
            self._import_descriptivenote(descriptivenote, rd_obj)
        rd_obj.full_clean()
        rd_obj.save()

    def _import_scopecontent(self, scopecontent, ead_obj):
        scopecontent_obj = ScopeContent(
            archdesc=ead_obj, scopecontent=serialise(scopecontent))
        import_attrs(scopecontent, SCOPECONTENT_ATTRS, scopecontent_obj)
        scopecontent_obj.full_clean()
        scopecontent_obj.save()

    def _import_script(self, script, languageset_obj):
        # script within languagedeclaration is handled within
        # _import_languagedeclaration.
        script_obj = LanguageSetScript(languageset=languageset_obj)
        import_attrs(script, SCRIPT_ATTRS, script_obj)
        scriptcode = script.get('scriptcode')
        if scriptcode:
            script_obj.scriptcode = search_term('iso15924', scriptcode)
        script_obj.script_el = script.text or ''
        script_obj.full_clean()
        script_obj.save()

    def _import_separatedmaterial(self, separatedmaterial, ead_obj):
        separatedmaterial_obj = SeparatedMaterial(
            archdesc=ead_obj, separatedmaterial=serialise(separatedmaterial))
        import_attrs(
            separatedmaterial, SEPARATEDMATERIAL_ATTRS, separatedmaterial_obj)
        separatedmaterial_obj.full_clean()
        separatedmaterial_obj.save()

    def _import_seriesstmt(self, seriesstmt, ead_obj):
        import_attrs(seriesstmt, SERIESSTMT_ATTRS, ead_obj, 'seriesstmt_')
        ead_obj.seriesstmt = serialise(seriesstmt)

    def _import_source(self, source, ead_obj):
        source_obj = Source(sources=ead_obj)
        import_attrs(source, SOURCE_ATTRS, source_obj)
        source_obj.save()
        for sourceentry in xpath(source, 'e:sourceentry'):
            self._import_sourceentry(sourceentry, source_obj)
        for objectxmlwrap in xpath(source, 'e:objectxmlwrap'):
            self._import_objectxmlwrap(objectxmlwrap, source_obj)
        for descriptivenote in xpath(source, 'e:descriptivenote'):
            self._import_descriptivenote(descriptivenote, source_obj)
        source_obj.full_clean()
        source_obj.save()

    def _import_sourceentry(self, sourceentry, source_obj):
        sourceentry_obj = SourceEntry(
            source=source_obj, sourceentry=sourceentry.text)
        import_attrs(sourceentry, SOURCEENTRY_ATTRS, sourceentry_obj)
        sourceentry_obj.full_clean()
        sourceentry_obj.save()

    def _import_sources(self, sources, ead_obj):
        import_attrs(sources, SOURCES_ATTRS, ead_obj, 'sources_')
        for source in sources:
            self._import_source(source, ead_obj)

    def _import_sponsor(self, sponsor, ead_obj):
        sponsor_obj = Sponsor(titlestmt=ead_obj, sponsor=serialise(sponsor))
        import_attrs(sponsor, SPONSOR_ATTRS, sponsor_obj)
        sponsor_obj.full_clean()
        sponsor_obj.save()

    def _import_subtitle(self, subtitle, ead_obj):
        subtitle_obj = Subtitle(
            titlestmt=ead_obj, subtitle=serialise(subtitle))
        import_attrs(subtitle, SUBTITLE_ATTRS, subtitle_obj)
        subtitle_obj.full_clean()
        subtitle_obj.save()

    def _import_term(self, term, localcontrol_obj):
        import_attrs(term, TERM_ATTRS, localcontrol_obj, 'term_')
        localcontrol_obj.term = term.text

    def _import_titleproper(self, titleproper, ead_obj):
        # Use only for titleproper within titlestmt; within
        # seriesstmt, no new model object is created.
        titleproper_obj = TitleProper(
            titlestmt=ead_obj, titleproper=serialise(titleproper))
        import_attrs(titleproper, TITLEPROPER_ATTRS, titleproper_obj)
        titleproper_obj.full_clean()
        titleproper_obj.save()

    def _import_titlestmt(self, titlestmt, ead_obj):
        import_attrs(titlestmt, TITLESTMT_ATTRS, ead_obj, 'titlestmt_')
        for titleproper in xpath(titlestmt, 'e:titleproper'):
            self._import_titleproper(titleproper, ead_obj)
        for subtitle in xpath(titlestmt, 'e:subtitle'):
            self._import_subtitle(subtitle, ead_obj)
        for author in xpath(titlestmt, 'e:author'):
            self._import_author(author, ead_obj)
        for sponsor in xpath(titlestmt, 'e:sponsor'):
            self._import_sponsor(sponsor, ead_obj)

    def _import_unitdate(self, unitdate, ead_obj):
        unitdate_obj = UnitDate(did=ead_obj, unitdate=serialise(unitdate))
        import_attrs(unitdate, UNITDATE_ATTRS, unitdate_obj)
        unitdate_obj.full_clean()
        unitdate_obj.save()

    def _import_unitdatestructured(self, uds, ead_obj):
        uds_obj = UnitDateStructured(did=ead_obj)
        import_attrs(uds, UNITDATESTRUCTURED_ATTRS, uds_obj)
        for dateset in xpath(uds, 'e:dateset'):
            import_attrs(dateset, DATESET_ATTRS, uds_obj, 'dateset_')
        uds_obj.full_clean()
        uds_obj.save()
        for daterange in xpath(uds, './/e:daterange'):
            self._import_daterange(
                daterange, uds_obj, UnitDateStructuredDateRange)
        for datesingle in xpath(uds, './/e:datesingle'):
            self._import_datesingle(
                datesingle, uds_obj, UnitDateStructuredDateSingle)

    def _import_unitid(self, unitid, ead_obj):
        unitid_obj = UnitId(did=ead_obj, unitid=serialise(unitid))
        import_attrs(unitid, UNITID_ATTRS, unitid_obj)
        unitid_obj.full_clean()
        unitid_obj.save()

    def _import_unittitle(self, unittitle, ead_obj):
        unittitle_obj = UnitTitle(did=ead_obj, unittitle=serialise(unittitle))
        import_attrs(unittitle, UNITTITLE_ATTRS, unittitle_obj)
        unittitle_obj.full_clean()
        unittitle_obj.save()

    def _import_unittype(self, unittype, physdescstructured_obj):
        physdescstructured_obj.unittype = unittype.text
        import_attrs(unittype, UNITTYPE_ATTRS, physdescstructured_obj,
                     'unittype_')

    def _import_userestrict(self, userestrict, ead_obj):
        userestrict_obj = UseRestrict(
            archdesc=ead_obj, userestrict=serialise(userestrict))
        import_attrs(userestrict, USERESTRICT_ATTRS, userestrict_obj)
        userestrict_obj.full_clean()
        userestrict_obj.save()



from lxml import etree

from .constants import (
    ABBR_ATTRS, ABSTRACT_ATTRS, ACCESSRESTRICT_ATTRS, ACCRUALS_ATTRS,
    ACQINFO_ATTRS, ADDRESS_ATTRS, ADDRESSLINE_ATTRS, AGENCYCODE_ATTRS,
    AGENCYNAME_ATTRS, AGENT_ATTRS, AGENTTYPE_ATTRS, ALTFORMAVAIL_ATTRS,
    APPRAISAL_ATTRS, ARCHDESC_ATTRS, ARRANGEMENT_ATTRS, AUTHOR_ATTRS,
    BIBLIOGRAPHY_ATTRS, BIOGHIST_ATTRS, CITATION_ATTRS, CONTAINER_ATTRS,
    CONTROL_ATTRS, CONTROLACCESS_ATTRS, CONTROLNOTE_ATTRS,
    CONVENTIONDECLARATION_ATTRS, CUSTODHIST_ATTRS, DAO_ATTRS, DAOSET_ATTRS,
    DATERANGE_ATTRS, DATESET_ATTRS, DATESINGLE_ATTRS, DESCRIPTIVENOTE_ATTRS,
    DID_ATTRS, DIDNOTE_ATTRS, DIMENSIONS_ATTRS, EAD_ATTRS, EAD_NAMESPACE,
    EAD_NS as EAD, EDITIONSTMT_ATTRS, EVENTDATETIME_ATTRS,
    EVENTDESCRIPTION_ATTRS, EVENTTYPE_ATTRS, FILEDESC_ATTRS, FILEPLAN_ATTRS,
    FROMTODATE_ATTRS, GEOGNAME_ATTRS, HEAD_ATTRS, INDEX_ATTRS,
    LANGMATERIAL_ATTRS, LANGUAGE_ATTRS, LANGUAGEDECLARATION_ATTRS,
    LANGUAGESET_ATTRS, LEGALSTATUS_ATTRS, LOCALCONTROL_ATTRS,
    LOCALTYPEDECLARATION_ATTRS, MAINTENANCEAGENCY_ATTRS,
    MAINTENANCEEVENT_ATTRS, MAINTENANCEHISTORY_ATTRS, MAINTENANCESTATUS_ATTRS,
    MATERIALSPEC_ATTRS, NAME_ATTRS, NOTESTMT_ATTRS, OBJECTXMLWRAP_ATTRS,
    ODD_ATTRS, ORIGINALSLOC_ATTRS, ORIGINATION_ATTRS, OTHERAGENCYCODE_ATTRS,
    OTHERFINDAID_ATTRS, PART_ATTRS, PHYSDESC_ATTRS, PHYSDESCSET_ATTRS,
    PHYSDESCSTRUCTURED_ATTRS, PHYSFACET_ATTRS, PHYSLOC_ATTRS, PHYSTECH_ATTRS,
    PREFERCITE_ATTRS, PROCESSINFO_ATTRS, PUBLICATIONSTATUS_ATTRS,
    PUBLICATIONSTMT_ATTRS, OTHERRECORDID_ATTRS, QUANTITY_ATTRS, RECORDID_ATTRS,
    RELATEDMATERIAL_ATTRS, RELATION_ATTRS, RELATIONENTRY_ATTRS,
    RELATIONS_ATTRS, REPOSITORY_ATTRS, REPRESENTATION_ATTRS,
    RIGHTSDECLARATION_ATTRS, SCOPECONTENT_ATTRS, SCRIPT_ATTRS,
    SEPARATEDMATERIAL_ATTRS, SERIESSTMT_ATTRS, SOURCE_ATTRS, SOURCEENTRY_ATTRS,
    SOURCES_ATTRS, SPONSOR_ATTRS, SUBTITLE_ATTRS, TERM_ATTRS,
    TITLEPROPER_ATTRS, TITLESTMT_ATTRS, UNITDATE_ATTRS,
    UNITDATESTRUCTURED_ATTRS, UNITID_ATTRS, UNITTITLE_ATTRS, UNITTYPE_ATTRS,
    USERESTRICT_ATTRS)


NS_MAP = {None: EAD_NAMESPACE}


def append_xml(parent, xml):
    if xml:
        root = etree.fromstring('<wrapper>{}</wrapper>'.format(xml))
        parent.text = root.text
        for child in root:
            parent.append(child)


def serialise_attrs(obj, attrs, element, prefix=''):
    for attr in attrs:
        if attr == 'id':
            attr = 'ead_id'
        value = getattr(obj, prefix + attr)
        if value is not None and value != '':
            if isinstance(value, list):
                value = value[0]
            if attr == 'ead_id':
                attr = 'id'
            elif attr in ('lang', 'script'):
                value = value.termid
            elif attr in ('approximate', 'parallel'):
                # BooleanFields.
                value = str(value).lower()
            element.set(attr, str(value))


class EADSerialiser:

    def serialise(self, ead_obj):
        ead = self.to_xml(ead_obj)
        return etree.tostring(ead, encoding='unicode', pretty_print=True)

    def to_xml(self, ead_obj):
        ead = etree.Element(EAD + 'ead', nsmap=NS_MAP)
        serialise_attrs(ead_obj, EAD_ATTRS, ead)
        self._serialise_control(ead_obj, ead)
        self._serialise_archdesc(ead_obj, ead)
        return ead

    def _serialise_abbr(self, parent_obj, parent):
        if not parent_obj.abbr:
            return
        abbr = etree.SubElement(parent, EAD + 'abbr')
        serialise_attrs(parent_obj, ABBR_ATTRS, abbr, 'abbr_')
        abbr.text = parent_obj.abbr

    def _serialise_abstract(self, abstract_obj, did):
        abstract = etree.SubElement(did, EAD + 'abstract')
        serialise_attrs(abstract_obj, ABSTRACT_ATTRS, abstract)
        append_xml(abstract, abstract_obj.abstract)

    def _serialise_accessrestrict(self, accessrestrict_obj, archdesc):
        accessrestrict = etree.SubElement(archdesc, EAD + 'accessrestrict')
        serialise_attrs(
            accessrestrict_obj, ACCESSRESTRICT_ATTRS, accessrestrict)
        append_xml(accessrestrict, accessrestrict_obj.accessrestrict)

    def _serialise_accruals(self, accruals_obj, archdesc):
        accruals = etree.SubElement(archdesc, EAD + 'accruals')
        serialise_attrs(accruals_obj, ACCRUALS_ATTRS, accruals)
        append_xml(accruals, accruals_obj.accruals)

    def _serialise_acqinfo(self, acqinfo_obj, archdesc):
        acqinfo = etree.SubElement(archdesc, EAD + 'acqinfo')
        serialise_attrs(acqinfo_obj, ACQINFO_ATTRS, acqinfo)
        append_xml(acqinfo, acqinfo_obj.acqinfo)

    def _serialise_addressline(self, addressline_obj, address):
        addressline = etree.SubElement(address, EAD + 'addressline')
        serialise_attrs(addressline_obj, ADDRESSLINE_ATTRS, addressline)
        append_xml(addressline, addressline_obj.addressline)

    def _serialise_agencycode(self, ead_obj, maintenanceagency):
        if ead_obj.agencycode:
            agencycode = etree.SubElement(
                maintenanceagency, EAD + 'agencycode')
            serialise_attrs(
                ead_obj, AGENCYCODE_ATTRS, agencycode, 'agencycode_')
            agencycode.text = ead_obj.agencycode

    def _serialise_agencyname(self, agencyname_obj, maintenanceagency):
        agencyname = etree.SubElement(maintenanceagency, EAD + 'agencyname')
        serialise_attrs(agencyname_obj, AGENCYNAME_ATTRS, agencyname)
        agencyname.text = agencyname_obj.agencyname

    def _serialise_agent(self, maintenanceevent_obj, maintenanceevent):
        agent = etree.SubElement(maintenanceevent, EAD + 'agent')
        serialise_attrs(maintenanceevent_obj, AGENT_ATTRS, agent, 'agent_')
        agent.text = maintenanceevent_obj.agent

    def _serialise_agenttype(self, maintenanceevent_obj, maintenanceevent):
        agenttype = etree.SubElement(maintenanceevent, EAD + 'agenttype')
        serialise_attrs(maintenanceevent_obj, AGENTTYPE_ATTRS, agenttype,
                        'agenttype_')
        agenttype.text = maintenanceevent_obj.agenttype

    def _serialise_altformavail(self, altformavail_obj, archdesc):
        altformavail = etree.SubElement(archdesc, EAD + 'altformavail')
        serialise_attrs(altformavail_obj, ALTFORMAVAIL_ATTRS, altformavail)
        append_xml(altformavail, altformavail_obj.altformavail)

    def _serialise_appraisal(self, appraisal_obj, archdesc):
        appraisal = etree.SubElement(archdesc, EAD + 'appraisal')
        serialise_attrs(appraisal_obj, APPRAISAL_ATTRS, appraisal)
        append_xml(appraisal, appraisal_obj.appraisal)

    def _serialise_archdesc(self, ead_obj, ead):
        archdesc = etree.SubElement(ead, EAD + 'archdesc')
        serialise_attrs(ead_obj, ARCHDESC_ATTRS, archdesc, 'archdesc_')
        self._serialise_did(ead_obj, archdesc)
        for accessrestrict_obj in ead_obj.accessrestrict_set.all():
            self._serialise_accessrestrict(accessrestrict_obj, archdesc)
        for accruals_obj in ead_obj.accruals_set.all():
            self._serialise_accruals(accruals_obj, archdesc)
        for acqinfo_obj in ead_obj.acqinfo_set.all():
            self._serialise_acqinfo(acqinfo_obj, archdesc)
        for altformavail_obj in ead_obj.altformavail_set.all():
            self._serialise_altformavail(altformavail_obj, archdesc)
        for appraisal_obj in ead_obj.appraisal_set.all():
            self._serialise_appraisal(appraisal_obj, archdesc)
        for arrangement_obj in ead_obj.arrangement_set.all():
            self._serialise_arrangement(arrangement_obj, archdesc)
        for bibliography_obj in ead_obj.bibliography_set.all():
            self._serialise_bibliography(bibliography_obj, archdesc)
        for bioghist_obj in ead_obj.bioghist_set.all():
            self._serialise_bioghist(bioghist_obj, archdesc)
        for controlaccess_obj in ead_obj.controlaccess_set.all():
            self._serialise_controlaccess(controlaccess_obj, archdesc)
        for custodhist_obj in ead_obj.custodhist_set.all():
            self._serialise_custodhist(custodhist_obj, archdesc)
        for fileplan_obj in ead_obj.fileplan_set.all():
            self._serialise_fileplan(fileplan_obj, archdesc)
        for index_obj in ead_obj.index_set.all():
            self._serialise_index(index_obj, archdesc)
        for legalstatus_obj in ead_obj.legalstatus_set.all():
            self._serialise_legalstatus(legalstatus_obj, archdesc)
        for odd_obj in ead_obj.odd_set.all():
            self._serialise_odd(odd_obj, archdesc)
        for originalsloc_obj in ead_obj.originalsloc_set.all():
            self._serialise_originalsloc(originalsloc_obj, archdesc)
        for otherfindaid_obj in ead_obj.otherfindaid_set.all():
            self._serialise_otherfindaid(otherfindaid_obj, archdesc)
        for phystech_obj in ead_obj.phystech_set.all():
            self._serialise_phystech(phystech_obj, archdesc)
        for prefercite_obj in ead_obj.prefercite_set.all():
            self._serialise_prefercite(prefercite_obj, archdesc)
        for processinfo_obj in ead_obj.processinfo_set.all():
            self._serialise_processinfo(processinfo_obj, archdesc)
        for relatedmaterial_obj in ead_obj.relatedmaterial_set.all():
            self._serialise_relatedmaterial(relatedmaterial_obj, archdesc)
        self._serialise_relations(ead_obj, archdesc)
        for scopecontent_obj in ead_obj.scopecontent_set.all():
            self._serialise_scopecontent(scopecontent_obj, archdesc)
        for separatedmaterial_obj in ead_obj.separatedmaterial_set.all():
            self._serialise_separatedmaterial(separatedmaterial_obj, archdesc)
        for userestrict_obj in ead_obj.userestrict_set.all():
            self._serialise_userestrict(userestrict_obj, archdesc)

    def _serialise_arrangement(self, arrangement_obj, archdesc):
        arrangement = etree.SubElement(archdesc, EAD + 'arrangement')
        serialise_attrs(arrangement_obj, ARRANGEMENT_ATTRS, arrangement)
        append_xml(arrangement, arrangement_obj.arrangement)

    def _serialise_author(self, author_obj, titlestmt):
        author = etree.SubElement(titlestmt, EAD + 'author')
        serialise_attrs(author_obj, AUTHOR_ATTRS, author)
        append_xml(author, author_obj.author)

    def _serialise_bibliography(self, bibliography_obj, archdesc):
        bibliography = etree.SubElement(archdesc, EAD + 'bibliography')
        serialise_attrs(bibliography_obj, BIBLIOGRAPHY_ATTRS, bibliography)
        append_xml(bibliography, bibliography_obj.bibliography)

    def _serialise_bioghist(self, bioghist_obj, archdesc):
        bioghist = etree.SubElement(archdesc, EAD + 'bioghist')
        serialise_attrs(bioghist_obj, BIOGHIST_ATTRS, bioghist)
        append_xml(bioghist, bioghist_obj.bioghist)

    def _serialise_citation(self, parent_obj, parent):
        citation = etree.SubElement(parent, EAD + 'citation')
        serialise_attrs(parent_obj, CITATION_ATTRS, citation, 'citation_')
        append_xml(citation, parent_obj.citation)

    def _serialise_container(self, container_obj, did):
        container = etree.SubElement(did, EAD + 'container')
        serialise_attrs(container_obj, CONTAINER_ATTRS, container)
        parents = [parent.ead_id for parent in container_obj.parents.all()]
        if parents:
            container.set('parent', ' '.join(parents))
        append_xml(container, container_obj.container)

    def _serialise_control(self, ead_obj, ead):
        control = etree.SubElement(ead, EAD + 'control')
        serialise_attrs(ead_obj, CONTROL_ATTRS, control, 'control_')
        self._serialise_recordid(ead_obj, control)
        for otherrecordid_obj in ead_obj.otherrecordid_set.all():
            self._serialise_otherrecordid(otherrecordid_obj, control)
        for representation_obj in ead_obj.representation_set.all():
            self._serialise_representation(representation_obj, control)
        self._serialise_filedesc(ead_obj, control)
        self._serialise_maintenancestatus(ead_obj, control)
        self._serialise_publicationstatus(ead_obj, control)
        self._serialise_maintenanceagency(ead_obj, control)
        for ld_obj in ead_obj.languagedeclaration_set.all():
            self._serialise_languagedeclaration(ld_obj, control)
        for cd_obj in ead_obj.conventiondeclaration_set.all():
            self._serialise_conventiondeclaration(cd_obj, control)
        for rd_obj in ead_obj.rightsdeclaration_set.all():
            self._serialise_rightsdeclaration(rd_obj, control)
        for ltd_obj in ead_obj.localtypedeclaration_set.all():
            self._serialise_localtypedeclaration(ltd_obj, control)
        for localcontrol_obj in ead_obj.localcontrol_set.all():
            self._serialise_localcontrol(localcontrol_obj, control)
        self._serialise_maintenancehistory(ead_obj, control)
        self._serialise_sources(ead_obj, control)

    def _serialise_controlaccess(self, controlaccess_obj, archdesc):
        controlaccess = etree.SubElement(archdesc, EAD + 'controlaccess')
        serialise_attrs(controlaccess_obj, CONTROLACCESS_ATTRS, controlaccess)
        append_xml(controlaccess, controlaccess_obj.controlaccess)

    def _serialise_controlnote(self, controlnote_obj, notestmt):
        controlnote = etree.SubElement(notestmt, EAD + 'controlnote')
        serialise_attrs(controlnote_obj, CONTROLNOTE_ATTRS, controlnote)
        append_xml(controlnote, controlnote_obj.controlnote)

    def _serialise_conventiondeclaration(self, conventiondeclaration_obj,
                                         control):
        conventiondeclaration = etree.SubElement(
            control, EAD + 'conventiondeclaration')
        serialise_attrs(conventiondeclaration_obj, CONVENTIONDECLARATION_ATTRS,
                        conventiondeclaration)
        self._serialise_abbr(conventiondeclaration_obj, conventiondeclaration)
        self._serialise_citation(
            conventiondeclaration_obj, conventiondeclaration)
        self._serialise_descriptivenote(
            conventiondeclaration_obj, conventiondeclaration)

    def _serialise_custodhist(self, custodhist_obj, archdesc):
        custodhist = etree.SubElement(archdesc, EAD + 'custodhist')
        serialise_attrs(custodhist_obj, CUSTODHIST_ATTRS, custodhist)
        append_xml(custodhist, custodhist_obj.custodhist)

    def _serialise_dao(self, dao_obj, parent):
        dao = etree.SubElement(parent, EAD + 'dao')
        serialise_attrs(dao_obj, DAO_ATTRS, dao)
        self._serialise_descriptivenote(dao_obj, dao)

    def _serialise_daoset(self, daoset_obj, did):
        daoset = etree.SubElement(did, EAD + 'daoset')
        serialise_attrs(daoset_obj, DAOSET_ATTRS, daoset)
        for dao_obj in daoset_obj.dao_set.all():
            self._serialise_dao(dao_obj, daoset)
        self._serialise_descriptivenote(daoset_obj, daoset)

    def _serialise_daterange(self, daterange_obj, parent, prefix=''):
        daterange = etree.SubElement(parent, EAD + 'daterange')
        serialise_attrs(daterange_obj, DATERANGE_ATTRS, daterange, prefix)
        if daterange_obj.fromdate or daterange_obj.fromdate_standarddate:
            self._serialise_fromtodate(daterange_obj, daterange, 'fromdate')
        if daterange_obj.todate or daterange_obj.todate_standarddate:
            self._serialise_fromtodate(daterange_obj, daterange, 'todate')

    def _serialise_dateset(self, obj, parent, daterange_objs, datesingle_objs):
        dateset = etree.SubElement(parent, EAD + 'dateset')
        serialise_attrs(obj, DATESET_ATTRS, dateset, 'dateset_')
        for daterange_obj in daterange_objs:
            self._serialise_daterange(daterange_obj, dateset)
        for datesingle_obj in datesingle_objs:
            self._serialise_datesingle(datesingle_obj, dateset)

    def _serialise_datesingle(self, datesingle_obj, parent, prefix=''):
        datesingle = etree.SubElement(parent, EAD + 'datesingle')
        serialise_attrs(datesingle_obj, DATESINGLE_ATTRS, datesingle, prefix)
        append_xml(datesingle, datesingle_obj.datesingle)

    def _serialise_descriptivenote(self, obj, parent, prefix=''):
        content = getattr(obj, prefix + 'descriptivenote')
        if not content:
            return
        descriptivenote = etree.SubElement(parent, EAD + 'descriptivenote')
        serialise_attrs(obj, DESCRIPTIVENOTE_ATTRS, descriptivenote,
                        prefix + 'descriptivenote_')
        append_xml(descriptivenote, content)

    def _serialise_did(self, ead_obj, archdesc):
        did = etree.SubElement(archdesc, EAD + 'did')
        serialise_attrs(ead_obj, DID_ATTRS, did, 'did_')
        self._serialise_head(ead_obj, did, 'did_')
        for abstract_obj in ead_obj.abstract_set.all():
            self._serialise_abstract(abstract_obj, did)
        for container_obj in ead_obj.container_set.all():
            self._serialise_container(container_obj, did)
        for dao_obj in ead_obj.dao_set.all():
            self._serialise_dao(dao_obj, did)
        for daoset_obj in ead_obj.daoset_set.all():
            self._serialise_daoset(daoset_obj, did)
        for didnote_obj in ead_obj.didnote_set.all():
            self._serialise_didnote(didnote_obj, did)
        for langmaterial_obj in ead_obj.langmaterial_set.all():
            self._serialise_langmaterial(langmaterial_obj, did)
        for materialspec_obj in ead_obj.materialspec_set.all():
            self._serialise_materialspec(materialspec_obj, did)
        for origination_obj in ead_obj.origination_set.all():
            self._serialise_origination(origination_obj, did)
        for physdesc_obj in ead_obj.physdesc_set.all():
            self._serialise_physdesc(physdesc_obj, did)
        for physdescset_obj in ead_obj.physdescset_set.all():
            self._serialise_physdescset(physdescset_obj, did)
        for physdescstructured_obj in ead_obj.physdescstructured_set.all():
            self._serialise_physdescstructured(physdescstructured_obj, did)
        for physloc_obj in ead_obj.physloc_set.all():
            self._serialise_physloc(physloc_obj, did)
        for repository_obj in ead_obj.repository_set.all():
            self._serialise_repository(repository_obj, did)
        for unitdate_obj in ead_obj.unitdate_set.all():
            self._serialise_unitdate(unitdate_obj, did)
        for unitdatestructured_obj in ead_obj.unitdatestructured_set.all():
            self._serialise_unitdatestructured(unitdatestructured_obj, did)
        for unitid_obj in ead_obj.unitid_set.all():
            self._serialise_unitid(unitid_obj, did)
        for unittitle_obj in ead_obj.unittitle_set.all():
            self._serialise_unittitle(unittitle_obj, did)

    def _serialise_didnote(self, didnote_obj, did):
        didnote = etree.SubElement(did, EAD + 'didnote')
        serialise_attrs(didnote_obj, DIDNOTE_ATTRS, didnote)
        append_xml(didnote, didnote_obj.didnote)

    def _serialise_dimensions(self, dimensions_obj, pds):
        dimensions = etree.SubElement(pds, EAD + 'dimensions')
        serialise_attrs(dimensions_obj, DIMENSIONS_ATTRS, dimensions)
        append_xml(dimensions, dimensions_obj.dimensions)

    def _serialise_editionstmt(self, ead_obj, filedesc):
        if not ead_obj.editionstmt:
            return
        editionstmt = etree.SubElement(filedesc, EAD + 'editionstmt')
        serialise_attrs(ead_obj, EDITIONSTMT_ATTRS, editionstmt,
                        'editionstmt_')
        append_xml(editionstmt, ead_obj.editionstmt)

    def _serialise_eventdatetime(self, maintenanceevent_obj, maintenanceevent):
        eventdatetime = etree.SubElement(
            maintenanceevent, EAD + 'eventdatetime')
        serialise_attrs(maintenanceevent_obj, EVENTDATETIME_ATTRS,
                        eventdatetime, 'eventdatetime_')
        eventdatetime.text = maintenanceevent_obj.eventdatetime

    def _serialise_eventdescription(self, ed_obj, maintenanceevent):
        eventdescription = etree.SubElement(
            maintenanceevent, EAD + 'eventdescription')
        serialise_attrs(ed_obj, EVENTDESCRIPTION_ATTRS, eventdescription)
        eventdescription.text = ed_obj.eventdescription

    def _serialise_eventtype(self, maintenanceevent_obj, maintenanceevent):
        eventtype = etree.SubElement(maintenanceevent, EAD + 'eventtype')
        serialise_attrs(maintenanceevent_obj, EVENTTYPE_ATTRS, eventtype,
                        'eventtype_')
        eventtype.text = maintenanceevent_obj.eventtype

    def _serialise_filedesc(self, ead_obj, control):
        filedesc = etree.SubElement(control, EAD + 'filedesc')
        serialise_attrs(ead_obj, FILEDESC_ATTRS, filedesc, 'filedesc_')
        self._serialise_titlestmt(ead_obj, filedesc)
        self._serialise_editionstmt(ead_obj, filedesc)
        self._serialise_publicationstmt(ead_obj, filedesc)
        self._serialise_seriesstmt(ead_obj, filedesc)
        self._serialise_notestmt(ead_obj, filedesc)

    def _serialise_fileplan(self, fileplan_obj, archdesc):
        fileplan = etree.SubElement(archdesc, EAD + 'fileplan')
        serialise_attrs(fileplan_obj, FILEPLAN_ATTRS, fileplan)
        append_xml(fileplan, fileplan_obj.fileplan)

    def _serialise_fromtodate(self, daterange_obj, daterange, element_name):
        date = etree.SubElement(daterange, EAD + element_name)
        # Use the individually stored attribute values for the
        # fromdate/todate elements, in place of those on the
        # serialised string.
        serialise_attrs(
            daterange_obj, FROMTODATE_ATTRS, date, element_name + '_')
        append_xml(date, getattr(daterange_obj, element_name))

    def _serialise_geogname(self, relation_obj, relation):
        if not relation_obj.geogname:
            return
        geogname = etree.SubElement(relation, EAD + 'geogname')
        serialise_attrs(relation_obj, GEOGNAME_ATTRS, geogname, 'geogname_')
        append_xml(geogname, relation_obj.geogname)

    def _serialise_head(self, obj, parent, prefix=''):
        content = getattr(obj, prefix + 'head')
        if not content:
            return
        head = etree.SubElement(parent, EAD + 'head')
        serialise_attrs(obj, HEAD_ATTRS, head, prefix + 'head_')
        append_xml(head, content)

    def _serialise_index(self, index_obj, archdesc):
        index = etree.SubElement(archdesc, EAD + 'index')
        serialise_attrs(index_obj, INDEX_ATTRS, index)
        append_xml(index, index_obj.index)

    def _serialise_langmaterial(self, langmaterial_obj, did):
        langmaterial = etree.SubElement(did, EAD + 'langmaterial')
        serialise_attrs(langmaterial_obj, LANGMATERIAL_ATTRS, langmaterial)
        for language_obj in langmaterial_obj.language_set.all():
            self._serialise_language(language_obj, langmaterial)
        for languageset_obj in langmaterial_obj.languageset_set.all():
            self._serialise_languageset(languageset_obj, langmaterial)
        self._serialise_descriptivenote(langmaterial_obj, langmaterial)

    def _serialise_language(self, obj, parent, prefix=''):
        langcode = getattr(obj, prefix + 'langcode')
        if not obj.language and not langcode:
            return
        language = etree.SubElement(parent, EAD + 'language')
        serialise_attrs(obj, LANGUAGE_ATTRS, language, prefix)
        if langcode:
            language.set('langcode', langcode.termid)
        language.text = obj.language

    def _serialise_languagedeclaration(self, ld_obj, control):
        ld = etree.SubElement(control, EAD + 'languagedeclaration')
        serialise_attrs(ld_obj, LANGUAGEDECLARATION_ATTRS, ld)
        self._serialise_language(ld_obj, ld, 'language_')
        self._serialise_script(ld_obj, ld, 'script_el_')
        self._serialise_descriptivenote(ld_obj, ld)

    def _serialise_languageset(self, languageset_obj, langmaterial):
        languageset = etree.SubElement(langmaterial, EAD + 'languageset')
        serialise_attrs(languageset_obj, LANGUAGESET_ATTRS, languageset)
        for language_obj in languageset_obj.language_set.all():
            self._serialise_language(language_obj, languageset)
        for script_obj in languageset_obj.script_set.all():
            self._serialise_script(script_obj, languageset)
        self._serialise_descriptivenote(languageset_obj, languageset)

    def _serialise_legalstatus(self, legalstatus_obj, archdesc):
        legalstatus = etree.SubElement(archdesc, EAD + 'legalstatus')
        serialise_attrs(legalstatus_obj, LEGALSTATUS_ATTRS, legalstatus)
        append_xml(legalstatus, legalstatus_obj.legalstatus)

    def _serialise_localcontrol(self, localcontrol_obj, control):
        localcontrol = etree.SubElement(control, EAD + 'localcontrol')
        serialise_attrs(localcontrol_obj, LOCALCONTROL_ATTRS, localcontrol)
        self._serialise_term(localcontrol_obj, localcontrol)
        if localcontrol_obj.fromdate or localcontrol_obj.todate or \
           localcontrol_obj.fromdate_standarddate or \
           localcontrol_obj.todate_standarddate:
            self._serialise_daterange(
                localcontrol_obj, localcontrol, 'daterange_')
        if localcontrol_obj.datesingle or \
           localcontrol_obj.datesingle_standarddate:
            self._serialise_datesingle(
                localcontrol_obj, localcontrol, 'datesingle_')

    def _serialise_localtypedeclaration(self, ltd_obj, control):
        ltd = etree.SubElement(control, EAD + 'localtypedeclaration')
        serialise_attrs(ltd_obj, LOCALTYPEDECLARATION_ATTRS, ltd)
        self._serialise_abbr(ltd_obj, ltd)
        self._serialise_citation(ltd_obj, ltd)
        self._serialise_descriptivenote(ltd_obj, ltd)

    def _serialise_maintenanceagency(self, ead_obj, control):
        maintenanceagency = etree.SubElement(
            control, EAD + 'maintenanceagency')
        serialise_attrs(ead_obj, MAINTENANCEAGENCY_ATTRS, maintenanceagency,
                        'maintenanceagency_')
        self._serialise_agencycode(ead_obj, maintenanceagency)
        for otheragencycode_obj in ead_obj.otheragencycode_set.all():
            self._serialise_otheragencycode(
                otheragencycode_obj, maintenanceagency)
        for agencyname_obj in ead_obj.agencyname_set.all():
            self._serialise_agencyname(agencyname_obj, maintenanceagency)
        self._serialise_descriptivenote(
            ead_obj, maintenanceagency, 'maintenanceagency_')

    def _serialise_maintenanceevent(self, maintenanceevent_obj,
                                    maintenancehistory):
        maintenanceevent = etree.SubElement(
            maintenancehistory, EAD + 'maintenanceevent')
        serialise_attrs(maintenanceevent_obj, MAINTENANCEEVENT_ATTRS,
                        maintenanceevent)
        self._serialise_eventtype(maintenanceevent_obj, maintenanceevent)
        self._serialise_eventdatetime(maintenanceevent_obj, maintenanceevent)
        self._serialise_agenttype(maintenanceevent_obj, maintenanceevent)
        self._serialise_agent(maintenanceevent_obj, maintenanceevent)
        for ed_obj in maintenanceevent_obj.eventdescription_set.all():
            self._serialise_eventdescription(ed_obj, maintenanceevent)

    def _serialise_maintenancehistory(self, ead_obj, control):
        maintenancehistory = etree.SubElement(
            control, EAD + 'maintenancehistory')
        serialise_attrs(ead_obj, MAINTENANCEHISTORY_ATTRS, maintenancehistory,
                        'maintenancehistory_')
        for maintenanceevent_obj in ead_obj.maintenanceevent_set.all():
            self._serialise_maintenanceevent(
                maintenanceevent_obj, maintenancehistory)

    def _serialise_maintenancestatus(self, ead_obj, control):
        maintenancestatus = etree.SubElement(
            control, EAD + 'maintenancestatus')
        serialise_attrs(ead_obj, MAINTENANCESTATUS_ATTRS, maintenancestatus,
                        'maintenancestatus_')
        if ead_obj.maintenancestatus:
            maintenancestatus.text = ead_obj.maintenancestatus

    def _serialise_materialspec(self, materialspec_obj, did):
        materialspec = etree.SubElement(did, EAD + 'materialspec')
        serialise_attrs(materialspec_obj, MATERIALSPEC_ATTRS, materialspec)
        append_xml(materialspec, materialspec_obj.materialspec)

    def _serialise_name(self, name_obj, parent, element_name):
        name = etree.SubElement(parent, EAD + element_name)
        serialise_attrs(name_obj, NAME_ATTRS, name)
        for part_obj in name_obj.part_set.all():
            self._serialise_part(part_obj, name)

    def _serialise_notestmt(self, ead_obj, filedesc):
        controlnote_objs = ead_obj.controlnote_set.all()
        if controlnote_objs:
            notestmt = etree.SubElement(filedesc, EAD + 'notestmt')
            serialise_attrs(ead_obj, NOTESTMT_ATTRS, notestmt, 'notestmt_')
            for controlnote_obj in controlnote_objs:
                self._serialise_controlnote(controlnote_obj, notestmt)

    def _serialise_objectxmlwrap(self, obj, parent):
        if not obj.objectxmlwrap:
            return
        objectxmlwrap = etree.SubElement(parent, EAD + 'objectxmlwrap')
        serialise_attrs(obj, OBJECTXMLWRAP_ATTRS, objectxmlwrap,
                        'objectxmlwrap_')
        append_xml(objectxmlwrap, obj.objectxmlwrap)

    def _serialise_odd(self, odd_obj, archdesc):
        odd = etree.SubElement(archdesc, EAD + 'odd')
        serialise_attrs(odd_obj, ODD_ATTRS, odd)
        append_xml(odd, odd_obj.odd)

    def _serialise_originalsloc(self, originalsloc_obj, archdesc):
        originalsloc = etree.SubElement(archdesc, EAD + 'originalsloc')
        serialise_attrs(originalsloc_obj, ORIGINALSLOC_ATTRS, originalsloc)
        append_xml(originalsloc, originalsloc_obj.originalsloc)

    def _serialise_origination(self, origination_obj, did):
        origination = etree.SubElement(did, EAD + 'origination')
        serialise_attrs(origination_obj, ORIGINATION_ATTRS, origination)
        for corpname_obj in origination_obj.corpname_set.all():
            self._serialise_name(corpname_obj, origination, 'corpname')
        for famname_obj in origination_obj.famname_set.all():
            self._serialise_name(famname_obj, origination, 'famname')
        for name_obj in origination_obj.name_set.all():
            self._serialise_name(name_obj, origination, 'name')
        for persname_obj in origination_obj.persname_set.all():
            self._serialise_name(persname_obj, origination, 'persname')

    def _serialise_otheragencycode(self, otheragencycode_obj,
                                   maintenanceagency):
        otheragencycode = etree.SubElement(
            maintenanceagency, EAD + 'otheragencycode')
        serialise_attrs(otheragencycode_obj, OTHERAGENCYCODE_ATTRS,
                        otheragencycode)
        otheragencycode.text = otheragencycode_obj.otheragencycode

    def _serialise_otherfindaid(self, otherfindaid_obj, archdesc):
        otherfindaid = etree.SubElement(archdesc, EAD + 'otherfindaid')
        serialise_attrs(otherfindaid_obj, OTHERFINDAID_ATTRS, otherfindaid)
        append_xml(otherfindaid, otherfindaid_obj.otherfindaid)

    def _serialise_otherrecordid(self, otherrecordid_obj, control):
        otherrecordid = etree.SubElement(control, EAD + 'otherrecordid')
        serialise_attrs(otherrecordid_obj, OTHERRECORDID_ATTRS, otherrecordid)
        otherrecordid.text = otherrecordid_obj.otherrecordid

    def _serialise_part(self, part_obj, name):
        part = etree.SubElement(name, EAD + 'part')
        serialise_attrs(part_obj, PART_ATTRS, part)
        append_xml(part, part_obj.part)

    def _serialise_physdesc(self, physdesc_obj, did):
        physdesc = etree.SubElement(did, EAD + 'physdesc')
        serialise_attrs(physdesc_obj, PHYSDESC_ATTRS, physdesc)
        append_xml(physdesc, physdesc_obj.physdesc)

    def _serialise_physdescset(self, physdescset_obj, did):
        physdescset = etree.SubElement(did, EAD + 'physdescset')
        serialise_attrs(physdescset_obj, PHYSDESCSET_ATTRS, physdescset)
        for pds_obj in physdescset_obj.physdescstructured_set.all():
            self._serialise_physdescstructured(pds_obj, physdescset)

    def _serialise_physdescstructured(self, physdescstructured_obj, parent):
        pds = etree.SubElement(parent, EAD + 'physdescstructured')
        serialise_attrs(physdescstructured_obj, PHYSDESCSTRUCTURED_ATTRS, pds)
        self._serialise_quantity(physdescstructured_obj, pds)
        self._serialise_unittype(physdescstructured_obj, pds)
        for dimensions_obj in physdescstructured_obj.dimensions_set.all():
            self._serialise_dimensions(dimensions_obj, pds)
        for physfacet_obj in physdescstructured_obj.physfacet_set.all():
            self._serialise_physfacet(physfacet_obj, pds)
        self._serialise_descriptivenote(physdescstructured_obj, pds)

    def _serialise_physfacet(self, physfacet_obj, physdescstructured):
        physfacet = etree.SubElement(physdescstructured, EAD + 'physfacet')
        serialise_attrs(physfacet_obj, PHYSFACET_ATTRS, physfacet)
        append_xml(physfacet, physfacet_obj.physfacet)

    def _serialise_physloc(self, physloc_obj, did):
        physloc = etree.SubElement(did, EAD + 'physloc')
        serialise_attrs(physloc_obj, PHYSLOC_ATTRS, physloc)
        parents = [parent.ead_id for parent in physloc_obj.parents.all()]
        if parents:
            physloc.set('parent', ' '.join(parents))
        append_xml(physloc, physloc_obj.physloc)

    def _serialise_phystech(self, phystech_obj, archdesc):
        phystech = etree.SubElement(archdesc, EAD + 'phystech')
        serialise_attrs(phystech_obj, PHYSTECH_ATTRS, phystech)
        append_xml(phystech, phystech_obj.phystech)

    def _serialise_prefercite(self, prefercite_obj, archdesc):
        prefercite = etree.SubElement(archdesc, EAD + 'prefercite')
        serialise_attrs(prefercite_obj, PREFERCITE_ATTRS, prefercite)
        append_xml(prefercite, prefercite_obj.prefercite)

    def _serialise_processinfo(self, processinfo_obj, archdesc):
        processinfo = etree.SubElement(archdesc, EAD + 'processinfo')
        serialise_attrs(processinfo_obj, PROCESSINFO_ATTRS, processinfo)
        append_xml(processinfo, processinfo_obj.processinfo)

    def _serialise_publicationstatus(self, ead_obj, control):
        status = ead_obj.publicationstatus_value
        if not status:
            return
        publicationstatus = etree.SubElement(
            control, EAD + 'publicationstatus')
        serialise_attrs(ead_obj, PUBLICATIONSTATUS_ATTRS, publicationstatus,
                        'publicationstatus_')
        publicationstatus.text = ead_obj.publicationstatus

    def _serialise_publicationstmt(self, ead_obj, filedesc):
        publicationstmt = etree.SubElement(filedesc, EAD + 'publicationstmt')
        serialise_attrs(ead_obj, PUBLICATIONSTMT_ATTRS, publicationstmt,
                        'publicationstmt_')
        append_xml(publicationstmt, ead_obj.publicationstmt)

    def _serialise_quantity(self, physdescstructured_obj, pds):
        if not physdescstructured_obj.quantity:
            return
        quantity = etree.SubElement(pds, EAD + 'quantity')
        serialise_attrs(
            physdescstructured_obj, QUANTITY_ATTRS, quantity, 'quantity_')
        quantity.text = str(physdescstructured_obj.quantity)

    def _serialise_recordid(self, ead_obj, control):
        recordid = etree.SubElement(control, EAD + 'recordid')
        serialise_attrs(ead_obj, RECORDID_ATTRS, recordid, 'recordid_')
        if ead_obj.recordid:
            recordid.text = ead_obj.recordid

    def _serialise_relatedmaterial(self, relatedmaterial_obj, archdesc):
        relatedmaterial = etree.SubElement(archdesc, EAD + 'relatedmaterial')
        serialise_attrs(
            relatedmaterial_obj, RELATEDMATERIAL_ATTRS, relatedmaterial)
        append_xml(relatedmaterial, relatedmaterial_obj.relatedmaterial)

    def _serialise_relation(self, relation_obj, relations):
        relation = etree.SubElement(relations, EAD + 'relation')
        serialise_attrs(relation_obj, RELATION_ATTRS, relation)
        for relationentry_obj in relation_obj.relationentry_set.all():
            self._serialise_relationentry(relationentry_obj, relation)
        self._serialise_objectxmlwrap(relation_obj, relation)
        daterange_objs = relation_obj.daterange_set.all()
        datesingle_objs = relation_obj.datesingle_set.all()
        if len(daterange_objs) + len(datesingle_objs) > 1:
            self._serialise_dateset(
                relation_obj, relation, daterange_objs, datesingle_objs)
        else:
            for daterange_obj in daterange_objs:
                self._serialise_daterange(daterange_obj, relation)
            for datesingle_obj in datesingle_objs:
                self._serialise_datesingle(datesingle_obj, relation)
        self._serialise_geogname(relation_obj, relation)
        self._serialise_descriptivenote(relation_obj, relation)

    def _serialise_relationentry(self, relationentry_obj, relation):
        relationentry = etree.SubElement(relation, EAD + 'relationentry')
        serialise_attrs(relationentry_obj, RELATIONENTRY_ATTRS, relationentry)
        relationentry.text = relationentry_obj.relationentry

    def _serialise_relations(self, ead_obj, archdesc):
        relation_objs = ead_obj.relation_set.all()
        if not relation_objs:
            return
        relations = etree.SubElement(archdesc, EAD + 'relations')
        serialise_attrs(ead_obj, RELATIONS_ATTRS, relations, 'relations_')
        for relation_obj in relation_objs:
            self._serialise_relation(relation_obj, relations)

    def _serialise_repository(self, repository_obj, did):
        repository = etree.SubElement(did, EAD + 'repository')
        serialise_attrs(repository_obj, REPOSITORY_ATTRS, repository)
        for corpname_obj in repository_obj.corpname_set.all():
            self._serialise_name(corpname_obj, repository, 'corpname')
        for famname_obj in repository_obj.famname_set.all():
            self._serialise_name(famname_obj, repository, 'famname')
        for name_obj in repository_obj.name_set.all():
            self._serialise_name(name_obj, repository, 'name')
        for persname_obj in repository_obj.persname_set.all():
            self._serialise_name(persname_obj, repository, 'persname')
        addressline_objs = repository_obj.addressline_set.all()
        if addressline_objs:
            address = etree.SubElement(repository, EAD + 'address')
            serialise_attrs(repository_obj, ADDRESS_ATTRS, address, 'address_')
            for addressline_obj in addressline_objs:
                self._serialise_addressline(addressline_obj, address)

    def _serialise_representation(self, representation_obj, control):
        representation = etree.SubElement(control, EAD + 'representation')
        serialise_attrs(representation_obj, REPRESENTATION_ATTRS,
                        representation)
        representation.text = representation_obj.representation

    def _serialise_rightsdeclaration(self, rightsdeclaration_obj, control):
        rightsdeclaration = etree.SubElement(
            control, EAD + 'rightsdeclaration')
        serialise_attrs(rightsdeclaration_obj, RIGHTSDECLARATION_ATTRS,
                        rightsdeclaration)
        self._serialise_abbr(rightsdeclaration_obj, rightsdeclaration)
        self._serialise_citation(rightsdeclaration_obj, rightsdeclaration)
        self._serialise_descriptivenote(
            rightsdeclaration_obj, rightsdeclaration)

    def _serialise_scopecontent(self, scopecontent_obj, parent):
        scopecontent = etree.SubElement(parent, EAD + 'scopecontent')
        serialise_attrs(scopecontent_obj, SCOPECONTENT_ATTRS, scopecontent)
        append_xml(scopecontent, scopecontent_obj.scopecontent)

    def _serialise_script(self, obj, parent, prefix=''):
        scriptcode = getattr(obj, prefix + 'scriptcode')
        if not obj.script_el and not scriptcode:
            return
        script = etree.SubElement(parent, EAD + 'script')
        serialise_attrs(obj, SCRIPT_ATTRS, script, prefix)
        if scriptcode:
            script.set('scriptcode', scriptcode.termid)
        script.text = obj.script_el

    def _serialise_separatedmaterial(self, sm_obj, archdesc):
        sm = etree.SubElement(archdesc, EAD + 'separatedmaterial')
        serialise_attrs(sm_obj, SEPARATEDMATERIAL_ATTRS, sm)
        append_xml(sm, sm_obj.separatedmaterial)

    def _serialise_seriesstmt(self, ead_obj, filedesc):
        if not ead_obj.seriesstmt:
            return
        seriesstmt = etree.SubElement(filedesc, EAD + 'seriesstmt')
        serialise_attrs(ead_obj, SERIESSTMT_ATTRS, seriesstmt, 'seriesstmt_')
        append_xml(seriesstmt, ead_obj.seriesstmt)

    def _serialise_source(self, source_obj, sources):
        source = etree.SubElement(sources, EAD + 'source')
        serialise_attrs(source_obj, SOURCE_ATTRS, source)
        for sourceentry_obj in source_obj.sourceentry_set.all():
            self._serialise_sourceentry(sourceentry_obj, source)
        self._serialise_objectxmlwrap(source_obj, source)
        self._serialise_descriptivenote(source_obj, source)

    def _serialise_sourceentry(self, sourceentry_obj, source):
        sourceentry = etree.SubElement(source, EAD + 'sourceentry')
        serialise_attrs(sourceentry_obj, SOURCEENTRY_ATTRS, sourceentry)
        sourceentry.text = sourceentry_obj.sourceentry

    def _serialise_sources(self, ead_obj, control):
        source_objs = ead_obj.source_set.all()
        if not source_objs:
            return
        sources = etree.SubElement(control, EAD + 'sources')
        serialise_attrs(ead_obj, SOURCES_ATTRS, sources, 'sources_')
        for source_obj in source_objs:
            self._serialise_source(source_obj, sources)

    def _serialise_sponsor(self, sponsor_obj, titlestmt):
        sponsor = etree.SubElement(titlestmt, EAD + 'sponsor')
        serialise_attrs(sponsor_obj, SPONSOR_ATTRS, sponsor)
        append_xml(sponsor, sponsor_obj.sponsor)

    def _serialise_subtitle(self, subtitle_obj, titlestmt):
        subtitle = etree.SubElement(titlestmt, EAD + 'subtitle')
        serialise_attrs(subtitle_obj, SUBTITLE_ATTRS, subtitle)
        append_xml(subtitle, subtitle_obj.subtitle)

    def _serialise_term(self, localcontrol_obj, localcontrol):
        if localcontrol_obj.term:
            term = etree.SubElement(localcontrol, EAD + 'term')
            serialise_attrs(localcontrol_obj, TERM_ATTRS, term, 'term_')
            term.text = localcontrol_obj.term

    def _serialise_titleproper(self, titleproper_obj, parent):
        titleproper = etree.SubElement(parent, EAD + 'titleproper')
        serialise_attrs(titleproper_obj, TITLEPROPER_ATTRS, titleproper)
        append_xml(titleproper, titleproper_obj.titleproper)

    def _serialise_titlestmt(self, ead_obj, filedesc):
        titlestmt = etree.SubElement(filedesc, EAD + 'titlestmt')
        serialise_attrs(ead_obj, TITLESTMT_ATTRS, titlestmt, 'titlestmt_')
        for titleproper_obj in ead_obj.titleproper_set.all():
            self._serialise_titleproper(titleproper_obj, titlestmt)
        for subtitle_obj in ead_obj.subtitle_set.all():
            self._serialise_subtitle(subtitle_obj, titlestmt)
        for author_obj in ead_obj.author_set.all():
            self._serialise_author(author_obj, titlestmt)
        for sponsor_obj in ead_obj.sponsor_set.all():
            self._serialise_sponsor(sponsor_obj, titlestmt)

    def _serialise_unitdate(self, unitdate_obj, did):
        unitdate = etree.SubElement(did, EAD + 'unitdate')
        serialise_attrs(unitdate_obj, UNITDATE_ATTRS, unitdate)
        append_xml(unitdate, unitdate_obj.unitdate)

    def _serialise_unitdatestructured(self, unitdatestructured_obj, did):
        unitdatestructured = etree.SubElement(did, EAD + 'unitdatestructured')
        serialise_attrs(unitdatestructured_obj, UNITDATESTRUCTURED_ATTRS,
                        unitdatestructured)
        daterange_objs = unitdatestructured_obj.daterange_set.all()
        datesingle_objs = unitdatestructured_obj.datesingle_set.all()
        if len(daterange_objs) + len(datesingle_objs) > 1:
            self._serialise_dateset(
                unitdatestructured_obj, unitdatestructured, daterange_objs,
                datesingle_objs)
        else:
            for daterange_obj in daterange_objs:
                self._serialise_daterange(daterange_obj, unitdatestructured)
            for datesingle_obj in datesingle_objs:
                self._serialise_datesingle(datesingle_obj, unitdatestructured)

    def _serialise_unitid(self, unitid_obj, did):
        unitid = etree.SubElement(did, EAD + 'unitid')
        serialise_attrs(unitid_obj, UNITID_ATTRS, unitid)
        append_xml(unitid, unitid_obj.unitid)

    def _serialise_unittitle(self, unittitle_obj, did):
        unittitle = etree.SubElement(did, EAD + 'unittitle')
        serialise_attrs(unittitle_obj, UNITTITLE_ATTRS, unittitle)
        append_xml(unittitle, unittitle_obj.unittitle)

    def _serialise_unittype(self, physdescstructured_obj, pds):
        if not physdescstructured_obj.unittype:
            return
        unittype = etree.SubElement(pds, EAD + 'unittype')
        serialise_attrs(
            physdescstructured_obj, UNITTYPE_ATTRS, unittype, 'unittype_')
        unittype.text = physdescstructured_obj.unittype

    def _serialise_userestrict(self, userestrict_obj, archdesc):
        userestrict = etree.SubElement(archdesc, EAD + 'userestrict')
        serialise_attrs(userestrict_obj, USERESTRICT_ATTRS, userestrict)
        append_xml(userestrict, userestrict_obj.userestrict)
